"""Processor for transforming analysis results to database format."""

from __future__ import annotations

import json
from collections.abc import Mapping
from datetime import datetime
from numbers import Number
from pathlib import Path
from typing import Any

import numpy as np

from metsuperq.data_handling.hdf5_data_manager import DEFAULT_METADATA_REQUIREMENTS
from metsuperq.database.schema import (
    FitQuality,
    InfluxPoint,
    MeasurementPoint,
    MeasurementTags,
    MeasurementType,
    T1MeasurementFields,
    T2EchoMeasurementFields,
    T2RamseyMeasurementFields,
)
from metsuperq.utils import setup_logging

logger = setup_logging(__name__)

# Type alias for analysis results from BaseAnalyzer
AnalysisResult = dict[str, Any]

# Constants for fit quality thresholds
EXCELLENT_R_SQUARED_THRESHOLD = 0.95
EXCELLENT_CHI_SQUARED_THRESHOLD = 2.0
GOOD_R_SQUARED_THRESHOLD = 0.90
GOOD_CHI_SQUARED_THRESHOLD = 5.0
FAIR_R_SQUARED_THRESHOLD = 0.80
FAIR_CHI_SQUARED_THRESHOLD = 10.0
POOR_R_SQUARED_THRESHOLD = 0.50

# Constants for TUID parsing
TUID_TIMESTAMP_LENGTH = 15
TUID_FIRST_DASH_POS = 8
TUID_SECOND_DASH_POS = 15


class MeasurementProcessor:
    """Transforms analysis results to InfluxDB measurement points."""

    def __init__(self) -> None:
        """Initialize measurement processor."""
        self._metadata_requirements = DEFAULT_METADATA_REQUIREMENTS.copy()
        logger.debug("Initialized MeasurementProcessor")

    def process_analysis_result(
        self,
        analysis_result: AnalysisResult,
        experiment_id: str,
        device_name: str,
        qubit_name: str,
    ) -> list[InfluxPoint]:
        """Transform analysis result to measurement points.

        Parameters
        ----------
        analysis_result : AnalysisResult
            Analysis result from BaseAnalyzer.run_full_analysis()
        experiment_id : str
            Experiment unique identifier
        device_name : str
            Name of the quantum device
        qubit_name : str
            Name of the measured qubit

        Returns
        -------
        list[InfluxPoint]
            List of measurement points with denormalized metadata
        """
        logger.info("Processing analysis result for qubit %s", qubit_name)

        # Check physics validation results
        physics_validation = analysis_result.get("physics_validation", {})
        if not physics_validation.get("is_valid", True):
            violations = physics_validation.get("violations", [])
            logger.error(
                "Physics validation FAILED for qubit %s. Recording anyway but flagging issues.",
                qubit_name,
            )
            for violation in violations:
                logger.error("  Physics violation: %s", violation)

        # Extract and validate metadata once (used for all measurements)
        metadata = analysis_result.get("metadata")
        if metadata is None:
            raise ValueError("Analysis result missing metadata; cannot create measurement points")

        validated_metadata = self._validate_and_normalize_metadata(metadata)

        points: list[InfluxPoint] = []
        fits = analysis_result.get("fits", {})
        timestamp = self._extract_timestamp(analysis_result)
        tuid = analysis_result.get("tuid")

        for measurement_type_str, fit_result in fits.items():
            if not fit_result.get("fit_success", False):
                logger.warning("Skipping failed fit for %s", measurement_type_str)
                continue

            try:
                point = self._create_measurement_point(
                    measurement_type_str=measurement_type_str,
                    fit_result=fit_result,
                    metadata=validated_metadata,
                    experiment_id=experiment_id,
                    device_name=device_name,
                    qubit_name=qubit_name,
                    timestamp=timestamp,
                    tuid=tuid,
                )
                points.append(point)
                logger.debug("Created measurement point for %s", measurement_type_str)

            except (ValueError, KeyError, TypeError) as e:
                logger.error(
                    "Failed to process %s measurement: %s", measurement_type_str, e, exc_info=True
                )
                continue

        logger.info("Processed %d measurement points with denormalized metadata", len(points))
        return points

    def _extract_timestamp(self, analysis_result: AnalysisResult) -> datetime:
        """Extract measurement timestamp from analysis result."""
        # Try to get timestamp from analysis result
        if "timestamp" in analysis_result:
            timestamp = analysis_result["timestamp"]
            if isinstance(timestamp, datetime):
                return timestamp
            elif isinstance(timestamp, str):
                return datetime.fromisoformat(timestamp.replace("Z", "+00:00"))

        # Try to extract from TUID (format: YYYYmmDD-HHMMSS-sss-******)
        tuid = analysis_result.get("tuid")
        if tuid:
            try:
                # Extract timestamp part (first 15 characters)
                timestamp_str = tuid[:15]  # YYYYmmDD-HHMMSS
                return datetime.strptime(timestamp_str, "%Y%m%d-%H%M%S")
            except ValueError:
                logger.warning("Could not parse timestamp from TUID: %s", tuid)

        # Try to extract from filepath
        filepath = analysis_result.get("filepath")
        if filepath:
            try:
                path = Path(filepath)
                # Look for TUID pattern in path components
                for part in path.parts:
                    if (
                        len(part) >= TUID_TIMESTAMP_LENGTH
                        and part[TUID_FIRST_DASH_POS] == "-"
                        and part[TUID_SECOND_DASH_POS] == "-"
                    ):
                        timestamp_str = part[:TUID_TIMESTAMP_LENGTH]
                        return datetime.strptime(timestamp_str, "%Y%m%d-%H%M%S")
            except (ValueError, IndexError, KeyError):
                pass  # Expected failures when parsing timestamp from filepath

        # Fallback to current time
        logger.warning("Could not extract timestamp, using current time")
        return datetime.now()

    def _create_measurement_point(  # noqa: PLR0913
        self,
        measurement_type_str: str,
        fit_result: dict[str, Any],
        metadata: dict[str, Any],
        experiment_id: str,
        device_name: str,
        qubit_name: str,
        timestamp: datetime,
        tuid: str | None,
    ) -> MeasurementPoint:
        """Create a measurement point with denormalized metadata."""
        # Map measurement type
        measurement_type = self._map_measurement_type(measurement_type_str)

        # Assess and log fit quality
        quality = self._assess_fit_quality(fit_result)
        if quality in [FitQuality.POOR, FitQuality.FAILED]:
            logger.warning(
                "Poor fit quality (%s) for %s: R²=%.3f, χ²=%.2f",
                quality.value,
                measurement_type_str,
                fit_result.get("r_squared", 0.0),
                fit_result.get("chi_squared", float("inf")),
            )
        else:
            logger.debug(
                "Fit quality (%s) for %s: R²=%.3f",
                quality.value,
                measurement_type_str,
                fit_result.get("r_squared", 0.0),
            )

        # Create tags with metadata identifiers
        tags = MeasurementTags(
            measurement_type=measurement_type,
            experiment_id=experiment_id,
            device_name=device_name,
            qubit_name=qubit_name,
            sample_identifier=metadata["sample_identifier"],
            qubit_type=metadata["qubit_type"],
            measurement_institute=metadata["measurement_institute"],
            measurement_fridge=metadata["measurement_fridge"],
        )

        # Create fields with denormalized metadata
        fields = self._create_measurement_fields(measurement_type, fit_result, metadata)

        # Create measurement name
        measurement_name = f"{measurement_type.value}"

        return MeasurementPoint(
            measurement=measurement_name,
            tags=tags,
            fields=fields,
            timestamp=timestamp,
            tuid=tuid,
        )

    def _validate_and_normalize_metadata(self, metadata: Mapping[str, Any]) -> dict[str, Any]:
        """Validate and normalize metadata for use in measurements.

        Returns
        -------
        dict[str, Any]
            Normalized metadata dictionary with all required fields

        Raises
        ------
        ValueError
            If mandatory metadata fields are missing
        """
        for required_key, is_mandatory in self._metadata_requirements.items():
            if is_mandatory and required_key not in metadata:
                raise ValueError(f"Missing mandatory metadata field '{required_key}'")

        normalized_fields: dict[str, Any] = {}
        for key, value in metadata.items():
            normalized_fields[key] = self._normalize_metadata_value(value)

        return normalized_fields

    @staticmethod
    def _normalize_metadata_value(value: Any) -> str | int | float | bool:
        """Normalize metadata values to Influx-friendly scalars."""
        # Handle numpy types first
        normalized = MeasurementProcessor._convert_numpy_types(value)

        # Handle collections (recursive)
        if isinstance(normalized, Mapping):
            return MeasurementProcessor._normalize_mapping(normalized)
        if isinstance(normalized, (list, tuple, set)):
            return MeasurementProcessor._normalize_sequence(normalized)

        # Convert special types to strings
        if isinstance(normalized, (datetime, Path)):
            normalized = (
                normalized.isoformat() if isinstance(normalized, datetime) else str(normalized)
            )

        # Handle primitives and numeric types
        if isinstance(normalized, (str, bool, int, float)):
            return normalized
        if isinstance(normalized, Number):
            return MeasurementProcessor._convert_to_int_or_float(normalized)

        return str(normalized)

    @staticmethod
    def _convert_numpy_types(value: Any) -> Any:
        """Convert numpy types to Python primitives."""
        if isinstance(value, np.generic):
            return value.item()
        if isinstance(value, np.ndarray):
            return value.tolist()
        return value

    @staticmethod
    def _normalize_mapping(mapping: Mapping[Any, Any]) -> str:
        """Normalize a mapping to JSON string."""
        normalized = {
            str(k): MeasurementProcessor._normalize_metadata_value(v) for k, v in mapping.items()
        }
        return json.dumps(normalized)

    @staticmethod
    def _normalize_sequence(sequence: list | tuple | set) -> str:
        """Normalize a sequence to JSON string."""
        items = [MeasurementProcessor._normalize_metadata_value(item) for item in sequence]
        return json.dumps(items)

    @staticmethod
    def _convert_to_int_or_float(value: Number) -> int | float:
        """Convert abstract Number type to concrete int or float.

        Prefers int if the value can be represented exactly as an integer.
        """
        try:
            return int(value)  # type: ignore[arg-type]
        except (ValueError, TypeError):
            return float(value)  # type: ignore[arg-type]

    @staticmethod
    def _parse_metadata_timestamp(value: str) -> datetime:
        """Parse timestamp from metadata, fallback to current time on failure."""
        try:
            return datetime.fromisoformat(value.replace("Z", "+00:00"))
        except ValueError:
            logger.warning("Invalid metadata timestamp '%s', using current time", value)
            return datetime.now()

    def _map_measurement_type(self, measurement_type_str: str) -> MeasurementType:
        """Map string measurement type to enum."""
        type_mapping = {
            "t1": MeasurementType.T1,
            "t2e": MeasurementType.T2_ECHO,
            "t2r": MeasurementType.T2_RAMSEY,
        }

        normalized = measurement_type_str.lower()
        if normalized not in type_mapping:
            raise ValueError(f"Unknown measurement type: {measurement_type_str}")

        return type_mapping[normalized]

    def _assess_fit_quality(self, fit_result: dict[str, Any]) -> FitQuality:
        """Assess fit quality based on fit statistics."""
        r_squared = fit_result.get("r_squared", 0.0)
        chi_squared = fit_result.get("chi_squared", float("inf"))

        # Simple quality assessment
        if (
            r_squared >= EXCELLENT_R_SQUARED_THRESHOLD
            and chi_squared < EXCELLENT_CHI_SQUARED_THRESHOLD
        ):
            return FitQuality.EXCELLENT
        elif r_squared >= GOOD_R_SQUARED_THRESHOLD and chi_squared < GOOD_CHI_SQUARED_THRESHOLD:
            return FitQuality.GOOD
        elif r_squared >= FAIR_R_SQUARED_THRESHOLD and chi_squared < FAIR_CHI_SQUARED_THRESHOLD:
            return FitQuality.FAIR
        elif r_squared >= POOR_R_SQUARED_THRESHOLD:
            return FitQuality.POOR
        else:
            return FitQuality.FAILED

    def _create_measurement_fields(
        self,
        measurement_type: MeasurementType,
        fit_result: dict[str, Any],
        metadata: dict[str, Any],
    ) -> T1MeasurementFields | T2EchoMeasurementFields | T2RamseyMeasurementFields:
        """Create measurement fields with denormalized metadata."""
        # Common physics results
        common_fields = {
            "decay_time": fit_result["decay_time"],
            "decay_time_stderr": fit_result["decay_time_stderr"],
            "amplitude": fit_result["amplitude"],
            "amplitude_stderr": fit_result["amplitude_stderr"],
            "r_squared": fit_result["r_squared"],
            "chi_squared": fit_result["chi_squared"],
        }

        # Add baseline if present
        if "baseline" in fit_result:
            common_fields["baseline"] = fit_result["baseline"]
        if "baseline_stderr" in fit_result:
            common_fields["baseline_stderr"] = fit_result["baseline_stderr"]

        # Common metadata fields (ALL fields denormalized for team visibility)
        # Note: metadata values are already normalized by _validate_and_normalize_metadata,
        # Explicit type conversions ensure pyright can verify types match Pydantic models

        # Helper to safely get optional string values
        def _opt_str(key: str) -> str | None:
            v = metadata.get(key)
            return str(v) if v is not None else None

        # Helper to safely get optional float values
        def _opt_float(key: str) -> float | None:
            v = metadata.get(key)
            return float(v) if v is not None else None

        # Helper to safely get optional bool values
        def _opt_bool(key: str) -> bool | None:
            v = metadata.get(key)
            return bool(v) if v is not None else None

        if measurement_type == MeasurementType.T1:
            return T1MeasurementFields(
                **common_fields,
                # Experimental conditions
                cooldown_days=int(metadata["time_since_cooldown_start_days"]),
                t1_averages=int(metadata["t1_averages"]),
                # Pulse parameters
                ctrlpulse_pi_duration_ns=int(metadata["ctrlpulse_pi_duration_ns"]),
                ctrlpulse_shape=str(metadata["ctrlpulse_shape"]),
                ctrlpulse_pi_amplitude=_opt_float("ctrlpulse_pi_amplitude"),
                readpulse_duration_ns=int(metadata["readpulse_duration_ns"]),
                readpulse_shape=str(metadata["readpulse_shape"]),
                readpulse_amplitude=float(metadata["readpulse_amplitude"]),
                reset_time_us=float(metadata["reset_time_us"]),
                # Qubit and resonator properties
                qubit_frequency=float(metadata["qubit_frequency"]),
                resonator_frequency=float(metadata["resonator_frequency"]),
                resonator_qtotal=float(metadata["resonator_qtotal"]),
                qubit_ej_ghz=_opt_float("qubit_ej_ghz"),
                qubit_ec_ghz=_opt_float("qubit_ec_ghz"),
                qubit_el_ghz=_opt_float("qubit_el_ghz"),
                anharmonicity_mhz=_opt_float("anharmonicity_mhz"),
                # Material properties
                sample_origin_cleanroom=str(metadata["sample_origin_cleanroom"]),
                material_substrate=str(metadata["material_substrate"]),
                material_superconductor=str(metadata["material_superconductor"]),
                material_jj=str(metadata["material_jj"]),
                fabrication_date=_opt_str("fabrication_date"),
                # Amplifier and temperature
                parametric_amplifier_used=str(metadata["parametric_amplifier_used"]),
                parametric_amplifier_gain_db=_opt_float("parametric_amplifier_gain_db"),
                base_temperature_mk=_opt_float("base_temperature_mk"),
                chip_package=_opt_str("chip_package"),
                # Wiring configuration
                control_line_number=str(metadata["control_line_number"]),
                control_line_config=str(metadata["control_line_config"]),
                readout_in_line_number=str(metadata["readout_in_line_number"]),
                readout_in_line_config=str(metadata["readout_in_line_config"]),
                readout_out_line_number=str(metadata["readout_out_line_number"]),
                readout_out_line_config=str(metadata["readout_out_line_config"]),
                fridge_wiring_reference=str(metadata["fridge_wiring_reference"]),
                # References and history
                data_reference=str(metadata["data_reference"]),
                tuneup_measurement_reference=str(metadata["tuneup_measurement_reference"]),
                sample_history_reference=_opt_str("sample_history_reference"),
                # Sweep parameters (optional)
                part_of_sweep=_opt_bool("part_of_sweep"),
                sweep_variable=_opt_str("sweep_variable"),
                sweep_variable_value=_opt_float("sweep_variable_value"),
                sweep_variable_unit=_opt_str("sweep_variable_unit"),
                sweep_variable_list=_opt_str("sweep_variable_list"),
                sweep_reference_next=_opt_str("sweep_reference_next"),
                sweep_reference_previous=_opt_str("sweep_reference_previous"),
            )

        if measurement_type == MeasurementType.T2_ECHO:
            return T2EchoMeasurementFields(
                **common_fields,
                # Experimental conditions
                cooldown_days=int(metadata["time_since_cooldown_start_days"]),
                t2e_averages=int(metadata["t2e_averages"]),
                # Pulse parameters
                ctrlpulse_pi_duration_ns=int(metadata["ctrlpulse_pi_duration_ns"]),
                ctrlpulse_shape=str(metadata["ctrlpulse_shape"]),
                ctrlpulse_pi_amplitude=_opt_float("ctrlpulse_pi_amplitude"),
                readpulse_duration_ns=int(metadata["readpulse_duration_ns"]),
                readpulse_shape=str(metadata["readpulse_shape"]),
                readpulse_amplitude=float(metadata["readpulse_amplitude"]),
                reset_time_us=float(metadata["reset_time_us"]),
                # Qubit and resonator properties
                qubit_frequency=float(metadata["qubit_frequency"]),
                resonator_frequency=float(metadata["resonator_frequency"]),
                resonator_qtotal=float(metadata["resonator_qtotal"]),
                qubit_ej_ghz=_opt_float("qubit_ej_ghz"),
                qubit_ec_ghz=_opt_float("qubit_ec_ghz"),
                qubit_el_ghz=_opt_float("qubit_el_ghz"),
                anharmonicity_mhz=_opt_float("anharmonicity_mhz"),
                # Material properties
                sample_origin_cleanroom=str(metadata["sample_origin_cleanroom"]),
                material_substrate=str(metadata["material_substrate"]),
                material_superconductor=str(metadata["material_superconductor"]),
                material_jj=str(metadata["material_jj"]),
                fabrication_date=_opt_str("fabrication_date"),
                # Amplifier and temperature
                parametric_amplifier_used=str(metadata["parametric_amplifier_used"]),
                parametric_amplifier_gain_db=_opt_float("parametric_amplifier_gain_db"),
                base_temperature_mk=_opt_float("base_temperature_mk"),
                chip_package=_opt_str("chip_package"),
                # Wiring configuration
                control_line_number=str(metadata["control_line_number"]),
                control_line_config=str(metadata["control_line_config"]),
                readout_in_line_number=str(metadata["readout_in_line_number"]),
                readout_in_line_config=str(metadata["readout_in_line_config"]),
                readout_out_line_number=str(metadata["readout_out_line_number"]),
                readout_out_line_config=str(metadata["readout_out_line_config"]),
                fridge_wiring_reference=str(metadata["fridge_wiring_reference"]),
                # References and history
                data_reference=str(metadata["data_reference"]),
                tuneup_measurement_reference=str(metadata["tuneup_measurement_reference"]),
                sample_history_reference=_opt_str("sample_history_reference"),
                # Sweep parameters (optional)
                part_of_sweep=_opt_bool("part_of_sweep"),
                sweep_variable=_opt_str("sweep_variable"),
                sweep_variable_value=_opt_float("sweep_variable_value"),
                sweep_variable_unit=_opt_str("sweep_variable_unit"),
                sweep_variable_list=_opt_str("sweep_variable_list"),
                sweep_reference_next=_opt_str("sweep_reference_next"),
                sweep_reference_previous=_opt_str("sweep_reference_previous"),
            )

        if measurement_type == MeasurementType.T2_RAMSEY:
            return T2RamseyMeasurementFields(
                **common_fields,
                # T2 Ramsey specific fields
                frequency=float(fit_result["frequency"]),
                frequency_stderr=float(fit_result["frequency_stderr"]),
                phase=float(fit_result["phase"]),
                phase_stderr=float(fit_result["phase_stderr"]),
                # Experimental conditions
                cooldown_days=int(metadata["time_since_cooldown_start_days"]),
                t2r_averages=int(metadata["t2r_averages"]),
                ramsey_detuning_hz=_opt_float("ramsey_detuning_hz"),
                # Pulse parameters
                ctrlpulse_pi_duration_ns=int(metadata["ctrlpulse_pi_duration_ns"]),
                ctrlpulse_shape=str(metadata["ctrlpulse_shape"]),
                ctrlpulse_pi_amplitude=_opt_float("ctrlpulse_pi_amplitude"),
                readpulse_duration_ns=int(metadata["readpulse_duration_ns"]),
                readpulse_shape=str(metadata["readpulse_shape"]),
                readpulse_amplitude=float(metadata["readpulse_amplitude"]),
                reset_time_us=float(metadata["reset_time_us"]),
                # Qubit and resonator properties
                qubit_frequency=float(metadata["qubit_frequency"]),
                resonator_frequency=float(metadata["resonator_frequency"]),
                resonator_qtotal=float(metadata["resonator_qtotal"]),
                qubit_ej_ghz=_opt_float("qubit_ej_ghz"),
                qubit_ec_ghz=_opt_float("qubit_ec_ghz"),
                qubit_el_ghz=_opt_float("qubit_el_ghz"),
                anharmonicity_mhz=_opt_float("anharmonicity_mhz"),
                # Material properties
                sample_origin_cleanroom=str(metadata["sample_origin_cleanroom"]),
                material_substrate=str(metadata["material_substrate"]),
                material_superconductor=str(metadata["material_superconductor"]),
                material_jj=str(metadata["material_jj"]),
                fabrication_date=_opt_str("fabrication_date"),
                # Amplifier and temperature
                parametric_amplifier_used=str(metadata["parametric_amplifier_used"]),
                parametric_amplifier_gain_db=_opt_float("parametric_amplifier_gain_db"),
                base_temperature_mk=_opt_float("base_temperature_mk"),
                chip_package=_opt_str("chip_package"),
                # Wiring configuration
                control_line_number=str(metadata["control_line_number"]),
                control_line_config=str(metadata["control_line_config"]),
                readout_in_line_number=str(metadata["readout_in_line_number"]),
                readout_in_line_config=str(metadata["readout_in_line_config"]),
                readout_out_line_number=str(metadata["readout_out_line_number"]),
                readout_out_line_config=str(metadata["readout_out_line_config"]),
                fridge_wiring_reference=str(metadata["fridge_wiring_reference"]),
                # References and history
                data_reference=str(metadata["data_reference"]),
                tuneup_measurement_reference=str(metadata["tuneup_measurement_reference"]),
                sample_history_reference=_opt_str("sample_history_reference"),
                # Sweep parameters (optional)
                part_of_sweep=_opt_bool("part_of_sweep"),
                sweep_variable=_opt_str("sweep_variable"),
                sweep_variable_value=_opt_float("sweep_variable_value"),
                sweep_variable_unit=_opt_str("sweep_variable_unit"),
                sweep_variable_list=_opt_str("sweep_variable_list"),
                sweep_reference_next=_opt_str("sweep_reference_next"),
                sweep_reference_previous=_opt_str("sweep_reference_previous"),
            )

        raise ValueError(f"Unsupported measurement type: {measurement_type}")



"""Physics validation for quantum measurement analysis.

This module enforces fundamental quantum physics constraints on coherence times
and other measurement parameters.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import numpy as np

from metsuperq.utils import setup_logging

logger = setup_logging(__name__)

# Type alias
AnalysisResult = dict[str, Any]

# Physics validation constants
MAX_COHERENCE_TIME_SECONDS = 3600  # 1 hour - catches unit conversion errors
MAX_RAMSEY_FREQUENCY_HZ = 10e6  # 10 MHz - typical upper limit
MIN_RAMSEY_FREQUENCY_HZ = 100  # 100 Hz - typical lower limit


class PhysicsViolationError(Exception):
    """Exception raised when measurement results violate fundamental physics laws.

    This indicates either:
    - Bad measurement data
    - Incorrect fitting
    - Hardware malfunction
    - Wrong measurement type assignment
    """

    def __init__(self, message: str, details: dict[str, Any] | None = None) -> None:
        """Initialize physics violation error.

        Parameters
        ----------
        message : str
            Error description
        details : dict[str, Any] | None
            Additional context about the violation
        """
        super().__init__(message)
        self.details = details or {}


@dataclass
class CoherenceTimeValidation:
    """Results of coherence time physics validation."""

    is_valid: bool
    violations: list[str]
    warnings: list[str]
    t1: float | None = None
    t2_echo: float | None = None
    t2_ramsey: float | None = None


def validate_coherence_times(  # noqa: C901
    analysis_result: AnalysisResult, strict: bool = False
) -> CoherenceTimeValidation:
    """Validate coherence times against fundamental physics constraints.

    Enforces the quantum physics constraint: T2 ≤ 2×T1

    This is a fundamental law for qubits:
    - T1: Energy relaxation time (population decay)
    - T2: Phase coherence time (dephasing)
    - T2*: Ramsey dephasing time (includes inhomogeneous broadening)

    The relationship T2 ≤ 2×T1 comes from the fact that energy relaxation
    (T1 process) also causes dephasing. In the absence of pure dephasing,
    T2 = 2×T1. Any pure dephasing reduces T2 below this limit.

    Parameters
    ----------
    analysis_result : AnalysisResult
        Analysis results containing derived quantities
    strict : bool, optional
        If True, raise exception on violation. If False, only log warnings.
        Default: False

    Returns
    -------
    CoherenceTimeValidation
        Validation results with violations and warnings

    Raises
    ------
    PhysicsViolationError
        If strict=True and physics constraints are violated
    """
    derived = analysis_result.get("derived_quantities", {})

    t1 = derived.get("T1")
    t2_echo = derived.get("T2_echo")
    t2_ramsey = derived.get("T2_ramsey")

    violations: list[str] = []
    warnings: list[str] = []

    # Validate T2_echo ≤ 2×T1
    if t1 is not None and t2_echo is not None:
        if t2_echo > 2 * t1:
            violation_msg = (
                f"PHYSICS VIOLATION: T2_echo ({t2_echo:.3e} s) > 2×T1 ({2 * t1:.3e} s). "
                f"This violates fundamental quantum mechanics! "
                f"Ratio: T2/2T1 = {t2_echo / (2 * t1):.2f}"
            )
            violations.append(violation_msg)
            logger.error(violation_msg)

        # Warn if T2 is suspiciously close to 2×T1 (might indicate fitting issues)
        elif t2_echo > 1.9 * t1:
            warning_msg = (
                f"T2_echo ({t2_echo:.3e} s) is very close to 2×T1 ({2 * t1:.3e} s). "
                f"Ratio: {t2_echo / (2 * t1):.2f}. This is unusual - check fit quality."
            )
            warnings.append(warning_msg)
            logger.warning(warning_msg)

    # Validate T2* ≤ T2_echo (Ramsey should be shorter than echo)
    if t2_ramsey is not None and t2_echo is not None:
        if t2_ramsey > t2_echo:
            # This is unusual but not strictly forbidden (can happen with specific noise)
            warning_msg = (
                f"T2* ({t2_ramsey:.3e} s) > T2_echo ({t2_echo:.3e} s). "
                f"This is unusual but possible with certain noise characteristics. "
                f"Verify measurement setup."
            )
            warnings.append(warning_msg)
            logger.warning(warning_msg)

    # Validate T2* ≤ 2×T1
    if t1 is not None and t2_ramsey is not None:
        if t2_ramsey > 2 * t1:
            violation_msg = (
                f"PHYSICS VIOLATION: T2* ({t2_ramsey:.3e} s) > 2×T1 ({2 * t1:.3e} s). "
                f"This violates fundamental quantum mechanics! "
                f"Ratio: T2*/2T1 = {t2_ramsey / (2 * t1):.2f}"
            )
            violations.append(violation_msg)
            logger.error(violation_msg)

    # Check coherence time validity in a single pass
    coherence_times = [("T1", t1), ("T2_echo", t2_echo), ("T2_ramsey", t2_ramsey)]
    for name, value in coherence_times:
        if value is None:
            continue

        # Check for negative or zero coherence times
        if value <= 0:
            violation_msg = (
                f"PHYSICS VIOLATION: {name} = {value:.3e} s ≤ 0. Coherence times must be positive!"
            )
            violations.append(violation_msg)
            logger.error(violation_msg)
        # Check for unreasonably large coherence times (> 1 hour) - catches unit conversion errors
        elif value > MAX_COHERENCE_TIME_SECONDS:
            warning_msg = (
                f"{name} = {value:.3e} s (> 1 hour). "
                f"This is extremely long for a qubit. Check units!"
            )
            warnings.append(warning_msg)
            logger.warning(warning_msg)

    validation = CoherenceTimeValidation(
        is_valid=len(violations) == 0,
        violations=violations,
        warnings=warnings,
        t1=t1,
        t2_echo=t2_echo,
        t2_ramsey=t2_ramsey,
    )

    # Raise exception if strict mode and violations found
    if strict and not validation.is_valid:
        raise PhysicsViolationError(
            f"Physics validation failed with {len(violations)} violation(s)",
            details={
                "violations": violations,
                "warnings": warnings,
                "t1": t1,
                "t2_echo": t2_echo,
                "t2_ramsey": t2_ramsey,
            },
        )

    return validation


def validate_ramsey_parameters(
    frequency: float, phase: float, amplitude: float, baseline: float
) -> list[str]:
    """Validate Ramsey fit parameters for physical reasonableness.

    Parameters
    ----------
    frequency : float
        Detuning frequency in Hz
    phase : float
        Phase in radians
    amplitude : float
        Oscillation amplitude
    baseline : float
        Baseline offset

    Returns
    -------
    list[str]
        List of warning messages (empty if all valid)
    """
    warnings: list[str] = []

    # Check frequency range (typical: 1 kHz - 1 MHz for superconducting qubits)
    if abs(frequency) > MAX_RAMSEY_FREQUENCY_HZ:
        warnings.append(
            f"Ramsey frequency {frequency:.1e} Hz is very large (> 10 MHz). "
            f"Check if this is expected for your system."
        )
    elif abs(frequency) < MIN_RAMSEY_FREQUENCY_HZ:
        warnings.append(
            f"Ramsey frequency {frequency:.1f} Hz is very small (< 100 Hz). "
            f"May be difficult to resolve within T2*."
        )

    # Check phase wrapping
    if abs(phase) > np.pi:
        warnings.append(
            f"Phase {phase:.2f} rad is outside [-π, π]. Consider wrapping to principal value."
        )

    # Check amplitude
    if amplitude < 0:
        warnings.append(f"Negative amplitude {amplitude:.3f}. This is unusual.")

    # Check baseline
    if baseline < 0:
        warnings.append(
            f"Negative baseline {baseline:.3f}. This might indicate readout calibration issues."
        )

    return warnings



"""Custom exceptions for database operations."""

from __future__ import annotations

from typing import Any


class DatabaseError(Exception):
    """Base exception for database operations."""

    def __init__(self, message: str, details: dict[str, Any] | None = None) -> None:
        super().__init__(message)
        self.details = details or {}


class InfluxDBConnectionError(DatabaseError):
    """Raised when unable to connect to InfluxDB."""

    pass


class WriteError(DatabaseError):
    """Raised when unable to write data to InfluxDB."""

    pass


class ValidationError(DatabaseError):
    """Raised when data validation fails."""

    pass


class ConfigurationError(DatabaseError):
    """Raised when configuration is invalid."""

    pass



"""NPL data converter module for QCoDeS database operations and data processing."""

import os
from typing import Any

import numpy as np
import qcodes.dataset as qd

from metsuperq.analysis.constants import MEASUREMENTS_PER_ITERATION
from metsuperq.utils import setup_logging

logger = setup_logging(__name__)

# Time conversion constants for NPL instrument data
# Instrument samples are in units that need conversion to physical time
SAMPLE_TO_MICROSECONDS = 40e-3  # Convert instrument samples to microseconds
SAMPLE_TO_SECONDS = 40e-9  # Convert instrument samples to seconds


# Known measurement type identifiers in NPL database
KNOWN_MEASUREMENT_TYPES = ["t1_data", "T2_Echo_data", "ramsey_data"]


def _extract_measurement_type_from_table(cursor: Any, table_name: str) -> str:
    """Extract measurement type identifier from database table parameters."""
    result_id = cursor.execute(
        "SELECT result_table_name,parameters FROM runs WHERE result_table_name=?", (table_name,)
    ).fetchall()

    try:
        parameters = result_id[0][1].split(",")  # split() already returns a list
    except (IndexError, AttributeError) as e:
        logger.warning(f"Failed to parse parameters for table {table_name}: {e}", exc_info=True)
        return "unknown_measurement"

    for item in parameters:
        if item in KNOWN_MEASUREMENT_TYPES:
            return item

    return "unknown_measurement"


def _fetch_iq_data(cursor: Any, table_name: str) -> list[complex]:
    """Fetch I and Q data from database and combine into complex numbers."""
    # Fetch I data
    cursor.execute("SELECT I FROM ?", (table_name,))
    i_data = [row[0] for row in cursor.fetchall() if isinstance(row[0], float)]

    # Fetch Q data
    cursor.execute("SELECT Q FROM ?", (table_name,))
    q_data = [row[0] for row in cursor.fetchall() if isinstance(row[0], float)]

    # Combine into complex numbers
    return [complex(i_val, q_val) for i_val, q_val in zip(i_data, q_data, strict=False)]


def _fetch_delay_data(cursor: Any, table_name: str) -> list[float]:
    """Fetch pulse delay data from database and convert to microseconds."""
    cursor.execute("SELECT pulse_delay FROM ?", (table_name,))
    delays = []

    for row in cursor.fetchall():
        if isinstance(row[0], (float, int, np.floating, np.integer)):
            # Convert from instrument sample to microseconds
            delays.append(float(np.round(row[0] * SAMPLE_TO_MICROSECONDS, 3)))

    return delays


def read_measurement_data_from_db(cursor: Any, table_entries: list[str]) -> dict[str, Any]:
    """Read quantum measurement data from database tables.

    Args:
        cursor: Database cursor object for executing SQL queries
        table_entries: List of table name entries to read data from

    Returns
    -------
        Dictionary containing organized measurement data with delays and complex IQ values
    """
    output_dict: dict[str, Any] = {}
    measurement_index = 0
    iteration_count = 0

    for table_entry in table_entries:
        table_name = table_entry[0]

        # Extract measurement type
        measurement_type = _extract_measurement_type_from_table(cursor, table_name)
        if measurement_type == "unknown_measurement":
            continue

        # Fetch data
        iq_data = _fetch_iq_data(cursor, table_name)
        delay_data = _fetch_delay_data(cursor, table_name)

        # Update output dictionary
        measurement_key = f"measurement_{measurement_index}"
        if measurement_key not in output_dict:
            logger.info("Added measurement index: %s", measurement_index)
            output_dict[measurement_key] = {}

        data_key = f"{measurement_type}_{measurement_index}"
        output_dict[measurement_key][data_key] = [delay_data, iq_data]
        iteration_count += 1

        if iteration_count == MEASUREMENTS_PER_ITERATION:
            measurement_index += 1
            iteration_count = 0

    return output_dict


# Backward compatibility alias
data_reader_from_db = read_measurement_data_from_db


def _add_datasets_to_output_dict(
    datasets: list, output_dict: dict[str, Any], measurement_type_prefix: str
) -> None:
    """Convert QCoDeS datasets to measurement dictionary format.

    Args:
        datasets: List of QCoDeS datasets to process
        output_dict: Dictionary to update with processed data
        measurement_type_prefix: Prefix for measurement type (e.g., "T1", "T2E", "T2R")
    """
    for index, dataset in enumerate(datasets):
        dataframe = dataset.to_pandas_dataframe()
        # Use .tolist() directly on numpy result - avoids intermediate array copy
        delay_times_s = (dataframe.pulse_delay.values * SAMPLE_TO_SECONDS).tolist()
        iq_values = (dataframe.I.values + 1j * dataframe.Q.values).tolist()
        timestamps = dataframe.time_stamp

        measurement_key = f"measurement_{index:04}"
        if measurement_key not in output_dict:
            output_dict[measurement_key] = {}

        data_key = f"{measurement_type_prefix}_{index:04}"
        output_dict[measurement_key][data_key] = {timestamps[0]: [delay_times_s, iq_values]}


def _load_experiment_datasets(experiment, experiment_name: str, start_offset: int, conn) -> list:
    """Load datasets for an experiment with interleaved run IDs.

    Args:
        experiment: Loaded QCoDeS experiment
        experiment_name: Name of the experiment for loading
        start_offset: Starting offset for run IDs (1 for T1, 2 for T2E, 3 for T2R)
        conn: Database connection

    Returns
    -------
        List of loaded datasets
    """
    datasets = []
    interleave_step = 3  # Experiments are interleaved with step of 3
    for i in range(start_offset, len(experiment) * interleave_step + 1, interleave_step):
        res_dataset = qd.load_by_run_spec(
            captured_run_id=i, experiment_name=experiment_name, conn=conn
        )
        datasets.append(res_dataset)
    return datasets


def read_from_qcodes_db(db_name: str) -> dict[str, Any]:
    """Read quantum measurement data from QCoDeS database.

    Args:
        db_name: Name of the database file

    Returns
    -------
        Dictionary containing organized measurement data from T1, T2Echo, and Ramsey experiments

    Note
    ----
        All database operations are performed within the connection scope.
        Data is extracted to Python objects before the connection is closed.
    """
    source_path = os.path.join(os.getcwd(), db_name)
    source_conn = qd.connect(source_path)

    try:
        # Load experiments and their datasets
        experiments = [
            ("T1", "T1", 1),
            ("T2Echo", "T2Echo", 2),
            ("ramsey_calibration", "ramsey_calibration", 3),
        ]
        prefixes = ["T1", "T2E", "T2R"]

        output_dict: dict[str, Any] = {}

        # All database access happens within this try block while connection is active
        for (exp_name, load_name, start_offset), prefix in zip(experiments, prefixes, strict=True):
            experiment = qd.load_experiment_by_name(name=exp_name, conn=source_conn)
            datasets = _load_experiment_datasets(experiment, load_name, start_offset, source_conn)
            # Data is converted to Python dicts/lists here, safe to use after connection closes
            _add_datasets_to_output_dict(datasets, output_dict, prefix)

        return output_dict
    finally:
        # Ensure connection is closed even if an exception occurs
        source_conn.close()
        logger.debug("Closed QCoDeS database connection: %s", source_path)


def convert_npl_to_hdf5_format(npl_data: dict[str, Any]) -> dict[str, Any]:
    """
    Convert NPL data format to HDF5DataManager expected format.

    Transforms:
        {measurement: {t_data: {timestamp: [delays, iq_vals]}}}
    To:
        {measurement: {t_data: [metadata, [[delays], [iq_vals]]]}}

    Parameters
    ----------
    npl_data : dict
        Data from read_from_qcodes_db()

    Returns
    -------
    dict
        Data in HDF5DataManager compatible format
    """
    logger.debug("Converting NPL data format to HDF5 format")
    converted = {}

    for measurement_iter, measurement_data in npl_data.items():
        converted[measurement_iter] = {}

        for t_data, time_data_dict in measurement_data.items():
            # Extract first (and typically only) timestamp entry using next()
            timestamp, data_arrays = next(iter(time_data_dict.items()))
            delays, iq_vals = data_arrays

            # Convert to expected HDF5 format: [metadata, [delays_array, iq_array]]
            converted[measurement_iter][t_data] = [
                timestamp,  # metadata (timestamp)
                [delays, iq_vals],  # [delays_array, iq_array]
            ]

    logger.debug("Converted %d measurements to HDF5 format", len(converted))
    return converted


def convert_from_dataset(long_interleaved_dataset: dict[str, Any]) -> dict[str, Any]:
    """Convert dataset format to organized measurement dictionary.

    Args:
        long_interleaved_dataset: Input dataset in interleaved format

    Returns
    -------
        Dictionary with converted measurement data structure
    """
    result_dict = {}

    for _measurement_iter, measurement_data in long_interleaved_dataset.items():
        iteration_dict = {}
        for measurement_type, type_data in measurement_data.items():
            timestamp = type_data[0]
            dataset = type_data[1].to_pandas_dataframe()
            delay_us = (dataset.pulse_delay.values * SAMPLE_TO_MICROSECONDS).tolist()
            # Vectorized complex number creation - much faster than loop
            iq_values = (dataset.I.values + 1j * dataset.Q.values).tolist()
            iteration_dict[measurement_type] = {timestamp: [delay_us, iq_values]}
        result_dict.update(iteration_dict)
    return result_dict



"""Integration layer for bridging analysis results to database."""

from __future__ import annotations

__all__ = [
    "record_measurement",
    "check_config",
    "MeasurementProcessor",
    "MeasurementRecorder",
    "InfluxConfig",
    "InfluxConfigError",
    "InfluxPoint",
]

import logging

from metsuperq.database.config import InfluxConfig, InfluxConfigError
from metsuperq.database.schema import InfluxPoint
from metsuperq.integrations.measurement_processor import MeasurementProcessor
from metsuperq.integrations.measurement_recorder import (
    MeasurementRecorder,
    record_measurement,
)

logger = logging.getLogger(__name__)


def check_config() -> bool:
    """Return True if InfluxDB config can be loaded from environment.

    Examples use this to provide a friendly readiness check.
    """
    try:
        InfluxConfig.from_env()
        logger.info("InfluxDB config loaded successfully")
        return True
    except InfluxConfigError:
        logger.error("Failed to load InfluxDB config")
        return False



"""Measurement recorder for data."""

from __future__ import annotations

from collections.abc import Callable
from contextlib import contextmanager
from typing import Any

from metsuperq.database.client import InfluxDBClient
from metsuperq.database.config import InfluxConfig
from metsuperq.database.exceptions import DatabaseError
from metsuperq.database.writer import MeasurementWriter
from metsuperq.integrations.measurement_processor import MeasurementProcessor
from metsuperq.utils import setup_logging

logger = setup_logging(__name__)

# Avoid magic values (ruff PLR2004)
VISIBLE_TOKEN_TAIL = 4

# Type alias for analysis results
AnalysisResult = dict[str, Any]


class MeasurementRecorder:
    """Recorder for measurements."""

    def __init__(
        self,
        config: InfluxConfig | None = None,
        *,
        config_loader: Callable[[], InfluxConfig] | None = None,
    ) -> None:
        """Initialize measurement recorder.

        Parameters
        ----------
        config : InfluxConfig | None, optional
            InfluxDB configuration. If None, loads from environment.
        """
        self._config: InfluxConfig | None = config
        self._config_loader: Callable[[], InfluxConfig] = config_loader or InfluxConfig.from_env
        self.processor = MeasurementProcessor()

        logger.info("Initialized MeasurementRecorder")
        if self._config is not None:
            self._log_configuration(self._config)
        else:
            logger.debug("Configuration load deferred until first usage")

    def record_measurement(
        self,
        analysis_result: AnalysisResult,
        experiment_id: str,
        device_name: str,
        qubit_name: str,
    ) -> None:
        """Record a single measurement to the database.

        Parameters
        ----------
        analysis_result : AnalysisResult
            Analysis result from BaseAnalyzer.run_full_analysis()
        experiment_id : str
            Experiment unique identifier
        device_name : str
            Name of the quantum device
        qubit_name : str
            Name of the measured qubit
        """
        logger.info("Recording measurement for qubit %s on device %s", qubit_name, device_name)

        try:
            # Process analysis result to measurement points
            points = self.processor.process_analysis_result(
                analysis_result=analysis_result,
                qubit_name=qubit_name,
                device_name=device_name,
                experiment_id=experiment_id,
            )

            if not points:
                logger.warning("No valid measurement points to record")
                return

            with self._get_writer() as writer:
                for point in points:
                    writer.write_point(point)

            logger.info("Successfully recorded %d measurement points", len(points))

        except Exception as e:
            logger.error("Failed to record measurement: %s", e, exc_info=True)
            raise DatabaseError(
                f"Failed to record measurement for {qubit_name}: {e}",
                details={
                    "qubit_name": qubit_name,
                    "device_name": device_name,
                    "error": str(e),
                },
            ) from e

    def health_check(self) -> dict[str, Any]:
        """Run a health check against the database."""
        try:
            with self._get_client() as client:
                return client.health_check()
        except Exception as e:
            logger.error("Health check failed: %s", e, exc_info=True)
            return {"status": "error", "healthy": False, "error": str(e)}

    @contextmanager
    def _get_client(self):
        """Context manager yielding a connected Influx client."""
        config = self._ensure_config()
        client = InfluxDBClient(config)
        try:
            client.connect()
            yield client
        finally:
            client.disconnect()

    @contextmanager
    def _get_writer(self):
        """Context manager yielding a writer bound to a client."""
        with self._get_client() as client:
            writer = MeasurementWriter(client)
            try:
                yield writer
            finally:
                close = getattr(writer, "close", None)
                if callable(close):
                    close()

    def _ensure_config(self) -> InfluxConfig:
        if self._config is None:
            logger.debug("Loading InfluxDB configuration via loader")
            self._config = self._config_loader()
            self._log_configuration(self._config)
        return self._config

    def _log_configuration(self, config: InfluxConfig) -> None:
        logger.debug(
            "Configuration: %s",
            {
                "host": config.host,
                "port": config.port,
                "token": (
                    f"***{config.token[-VISIBLE_TOKEN_TAIL:]}"
                    if len(config.token) >= VISIBLE_TOKEN_TAIL
                    else "***"
                ),
                "org": config.org,
                "bucket": config.bucket,
                "timeout": config.timeout,
                "verify_ssl": config.verify_ssl,
            },
        )


def get_default_recorder() -> MeasurementRecorder:
    """Get or create the default measurement recorder singleton.

    Implemented via a function attribute to avoid ``global`` (ruff PLW0603).
    Tests may reset this by deleting the attribute.
    """
    if not hasattr(get_default_recorder, "_instance"):
        get_default_recorder._instance = MeasurementRecorder()  # type: ignore[attr-defined]
    return get_default_recorder._instance  # type: ignore[no-any-return]


def record_measurement(
    analysis_result: AnalysisResult,
    experiment_id: str,
    device_name: str,
    qubit_name: str,
) -> None:
    """Record a single measurement to the database.

    Parameters
    ----------
    analysis_result : AnalysisResult
        Analysis result from BaseAnalyzer.run_full_analysis()
    experiment_id : str
        Experiment unique identifier
    device_name : str
        Name of the quantum device (e.g., "device_1")
    qubit_name : str
        Name of the measured qubit (e.g., "Q1", "qubit_01")

    Examples
    --------
    >>> from metsuperq.analysis.base_analysis import BaseAnalyzer
    >>> from metsuperq.integrations import record_measurement
    >>>
    >>> analyzer = BaseAnalyzer()
    >>> result = analyzer.run_full_analysis("20250903-173311-935-8d9a03")
    >>> record_measurement(result, "Q1", "my_device")
    """
    recorder = get_default_recorder()
    recorder.record_measurement(
        analysis_result=analysis_result,
        experiment_id=experiment_id,
        device_name=device_name,
        qubit_name=qubit_name,
    )



"""Pydantic models for quantum measurement data schema."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Protocol

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from metsuperq.utils import setup_logging

logger = setup_logging(__name__)


class MeasurementType(str, Enum):
    """Supported quantum measurement types."""

    T1 = "t1"
    T2_ECHO = "t2e"
    T2_RAMSEY = "t2r"


class FitQuality(str, Enum):
    """Fit quality assessment."""

    EXCELLENT = "excellent"
    GOOD = "good"
    FAIR = "fair"
    POOR = "poor"
    FAILED = "failed"


class InfluxPoint(Protocol):
    """Protocol for models that can be written to InfluxDB."""

    measurement: str

    def to_influx_dict(self) -> dict[str, Any]:
        """Convert point to InfluxDB-compatible dictionary."""
        ...

    def model_dump(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Dump model representation."""
        ...


class MeasurementTags(BaseModel):
    """Tags for categorizing measurements (indexed in InfluxDB).

    Tags are indexed and used for filtering/grouping. Keep cardinality low.
    Use tags for: identifiers, categorical metadata, discrete values.
    """

    measurement_type: MeasurementType = Field(description="Type of measurement")
    experiment_id: str = Field(description="Experiment unique identifier")
    device_name: str = Field(description="Quantum device name")
    qubit_name: str = Field(description="Qubit name")

    # Core experimental context (low cardinality, used for filtering)
    sample_identifier: str = Field(description="Sample/chip identifier")
    qubit_type: str = Field(description="Qubit type (e.g., transmon, fluxonium)")
    measurement_institute: str = Field(description="Institution performing measurement")
    measurement_fridge: str = Field(description="Dilution fridge identifier")

    @field_validator("experiment_id", "device_name", "qubit_name", "sample_identifier")
    @classmethod
    def validate_identifiers(cls, v: str) -> str:
        """Validate identifier format (no spaces, special chars)."""
        if not v.replace("_", "").replace("-", "").isalnum():
            raise ValueError("Identifiers must be alphanumeric with _ or - only")
        return v


class T1MeasurementFields(BaseModel):
    """Fields for T1 relaxation measurements.

    Fields store measured values and high-cardinality context.
    Not indexed, but queryable and used for calculations.
    """

    # Physics results
    decay_time: float = Field(description="T1 decay time in seconds", gt=0, allow_inf_nan=False)
    decay_time_stderr: float = Field(
        description="Standard error of decay time", ge=0, allow_inf_nan=False
    )
    amplitude: float = Field(description="Fit amplitude", ge=0, allow_inf_nan=False)
    amplitude_stderr: float = Field(
        description="Standard error of amplitude", ge=0, allow_inf_nan=False
    )
    baseline: float | None = Field(default=None, description="Baseline offset", allow_inf_nan=False)
    baseline_stderr: float | None = Field(
        default=None, description="Baseline stderr", ge=0, allow_inf_nan=False
    )
    r_squared: float = Field(description="Fit R-squared", ge=0, le=1, allow_inf_nan=False)
    chi_squared: float = Field(description="Chi-squared statistic", ge=0, allow_inf_nan=False)

    # Experimental conditions (denormalized from metadata)
    cooldown_days: int = Field(description="Days since cooldown start")
    t1_averages: int = Field(description="Number of averages for T1 measurement", gt=0)

    # Pulse parameters
    ctrlpulse_pi_duration_ns: int = Field(description="Control pulse pi duration in ns", gt=0)
    ctrlpulse_shape: str = Field(description="Control pulse shape")
    ctrlpulse_pi_amplitude: float | None = Field(
        default=None, description="Control pulse pi amplitude"
    )
    readpulse_duration_ns: int = Field(description="Readout pulse duration in ns", gt=0)
    readpulse_shape: str = Field(description="Readout pulse shape")
    readpulse_amplitude: float = Field(description="Readout pulse amplitude")
    reset_time_us: float = Field(description="Reset time in microseconds")

    # Qubit and resonator properties
    qubit_frequency: float = Field(description="Qubit frequency in Hz", gt=0)
    resonator_frequency: float = Field(description="Resonator frequency in Hz", gt=0)
    resonator_qtotal: float = Field(description="Resonator total Q factor")
    qubit_ej_ghz: float | None = Field(default=None, description="Qubit EJ in GHz")
    qubit_ec_ghz: float | None = Field(default=None, description="Qubit EC in GHz")
    qubit_el_ghz: float | None = Field(default=None, description="Qubit EL in GHz")
    anharmonicity_mhz: float | None = Field(default=None, description="Anharmonicity in MHz")

    # Material properties
    sample_origin_cleanroom: str = Field(description="Sample origin cleanroom")
    material_substrate: str = Field(description="Substrate material")
    material_superconductor: str = Field(description="Superconductor material")
    material_jj: str = Field(description="Josephson junction material")
    fabrication_date: str | None = Field(default=None, description="Fabrication date")

    # Amplifier and temperature
    parametric_amplifier_used: str = Field(description="Parametric amplifier identifier")
    parametric_amplifier_gain_db: float | None = Field(
        default=None, description="Amplifier gain in dB"
    )
    base_temperature_mk: float | None = Field(default=None, description="Base temperature in mK")
    chip_package: str | None = Field(default=None, description="Chip package type")

    # Wiring configuration
    control_line_number: str = Field(description="Control line number")
    control_line_config: str = Field(description="Control line configuration")
    readout_in_line_number: str = Field(description="Readout input line number")
    readout_in_line_config: str = Field(description="Readout input line configuration")
    readout_out_line_number: str = Field(description="Readout output line number")
    readout_out_line_config: str = Field(description="Readout output line configuration")
    fridge_wiring_reference: str = Field(description="Fridge wiring reference")

    # References and history
    data_reference: str = Field(description="Data reference")
    tuneup_measurement_reference: str = Field(description="Tuneup measurement reference")
    sample_history_reference: str | None = Field(
        default=None, description="Sample history reference"
    )

    # Sweep parameters (optional)
    part_of_sweep: bool | None = Field(default=None, description="Whether part of a sweep")
    sweep_variable: str | None = Field(default=None, description="Sweep variable name")
    sweep_variable_value: float | None = Field(default=None, description="Sweep variable value")
    sweep_variable_unit: str | None = Field(default=None, description="Sweep variable unit")
    sweep_variable_list: str | None = Field(default=None, description="Sweep variable list (JSON)")
    sweep_reference_next: str | None = Field(default=None, description="Next sweep reference")
    sweep_reference_previous: str | None = Field(
        default=None, description="Previous sweep reference"
    )

    @model_validator(mode="after")
    def validate_errors(self) -> T1MeasurementFields:
        """Validate error values make sense."""
        if self.decay_time_stderr > self.decay_time:
            logger.warning("Decay time error larger than value itself")
        return self


class T2EchoMeasurementFields(BaseModel):
    """Fields for T2 echo measurements."""

    # Physics results
    decay_time: float = Field(description="T2 echo time in seconds", gt=0, allow_inf_nan=False)
    decay_time_stderr: float = Field(
        description="Standard error of decay time", ge=0, allow_inf_nan=False
    )
    amplitude: float = Field(description="Fit amplitude", ge=0, allow_inf_nan=False)
    amplitude_stderr: float = Field(
        description="Standard error of amplitude", ge=0, allow_inf_nan=False
    )
    baseline: float | None = Field(default=None, description="Baseline offset", allow_inf_nan=False)
    baseline_stderr: float | None = Field(
        default=None, description="Baseline stderr", ge=0, allow_inf_nan=False
    )
    r_squared: float = Field(description="Fit R-squared", ge=0, le=1, allow_inf_nan=False)
    chi_squared: float = Field(description="Chi-squared statistic", ge=0, allow_inf_nan=False)

    # Experimental conditions (denormalized from metadata)
    cooldown_days: int = Field(description="Days since cooldown start")
    t2e_averages: int = Field(description="Number of averages for T2 echo measurement", gt=0)

    # Pulse parameters
    ctrlpulse_pi_duration_ns: int = Field(description="Control pulse pi duration in ns", gt=0)
    ctrlpulse_shape: str = Field(description="Control pulse shape")
    ctrlpulse_pi_amplitude: float | None = Field(
        default=None, description="Control pulse pi amplitude"
    )
    readpulse_duration_ns: int = Field(description="Readout pulse duration in ns", gt=0)
    readpulse_shape: str = Field(description="Readout pulse shape")
    readpulse_amplitude: float = Field(description="Readout pulse amplitude")
    reset_time_us: float = Field(description="Reset time in microseconds")

    # Qubit and resonator properties
    qubit_frequency: float = Field(description="Qubit frequency in Hz", gt=0)
    resonator_frequency: float = Field(description="Resonator frequency in Hz", gt=0)
    resonator_qtotal: float = Field(description="Resonator total Q factor")
    qubit_ej_ghz: float | None = Field(default=None, description="Qubit EJ in GHz")
    qubit_ec_ghz: float | None = Field(default=None, description="Qubit EC in GHz")
    qubit_el_ghz: float | None = Field(default=None, description="Qubit EL in GHz")
    anharmonicity_mhz: float | None = Field(default=None, description="Anharmonicity in MHz")

    # Material properties
    sample_origin_cleanroom: str = Field(description="Sample origin cleanroom")
    material_substrate: str = Field(description="Substrate material")
    material_superconductor: str = Field(description="Superconductor material")
    material_jj: str = Field(description="Josephson junction material")
    fabrication_date: str | None = Field(default=None, description="Fabrication date")

    # Amplifier and temperature
    parametric_amplifier_used: str = Field(description="Parametric amplifier identifier")
    parametric_amplifier_gain_db: float | None = Field(
        default=None, description="Amplifier gain in dB"
    )
    base_temperature_mk: float | None = Field(default=None, description="Base temperature in mK")
    chip_package: str | None = Field(default=None, description="Chip package type")

    # Wiring configuration
    control_line_number: str = Field(description="Control line number")
    control_line_config: str = Field(description="Control line configuration")
    readout_in_line_number: str = Field(description="Readout input line number")
    readout_in_line_config: str = Field(description="Readout input line configuration")
    readout_out_line_number: str = Field(description="Readout output line number")
    readout_out_line_config: str = Field(description="Readout output line configuration")
    fridge_wiring_reference: str = Field(description="Fridge wiring reference")

    # References and history
    data_reference: str = Field(description="Data reference")
    tuneup_measurement_reference: str = Field(description="Tuneup measurement reference")
    sample_history_reference: str | None = Field(
        default=None, description="Sample history reference"
    )

    # Sweep parameters (optional)
    part_of_sweep: bool | None = Field(default=None, description="Whether part of a sweep")
    sweep_variable: str | None = Field(default=None, description="Sweep variable name")
    sweep_variable_value: float | None = Field(default=None, description="Sweep variable value")
    sweep_variable_unit: str | None = Field(default=None, description="Sweep variable unit")
    sweep_variable_list: str | None = Field(default=None, description="Sweep variable list (JSON)")
    sweep_reference_next: str | None = Field(default=None, description="Next sweep reference")
    sweep_reference_previous: str | None = Field(
        default=None, description="Previous sweep reference"
    )


class T2RamseyMeasurementFields(BaseModel):
    """Fields for T2 Ramsey measurements."""

    # Physics results
    decay_time: float = Field(description="T2* decay time in seconds", gt=0, allow_inf_nan=False)
    decay_time_stderr: float = Field(
        description="Standard error of decay time", ge=0, allow_inf_nan=False
    )
    amplitude: float = Field(description="Fit amplitude", ge=0, allow_inf_nan=False)
    amplitude_stderr: float = Field(
        description="Standard error of amplitude", ge=0, allow_inf_nan=False
    )
    frequency: float = Field(description="Ramsey oscillation frequency in Hz", allow_inf_nan=False)
    frequency_stderr: float = Field(
        description="Standard error of frequency", ge=0, allow_inf_nan=False
    )
    phase: float = Field(description="Phase in radians", allow_inf_nan=False)
    phase_stderr: float = Field(description="Standard error of phase", ge=0, allow_inf_nan=False)
    baseline: float | None = Field(default=None, description="Baseline offset", allow_inf_nan=False)
    baseline_stderr: float | None = Field(
        default=None, description="Baseline stderr", ge=0, allow_inf_nan=False
    )
    r_squared: float = Field(description="Fit R-squared", ge=0, le=1, allow_inf_nan=False)
    chi_squared: float = Field(description="Chi-squared statistic", ge=0, allow_inf_nan=False)

    # Experimental conditions (denormalized from metadata)
    cooldown_days: int = Field(description="Days since cooldown start")
    t2r_averages: int = Field(description="Number of averages for T2 Ramsey measurement", gt=0)
    ramsey_detuning_hz: float | None = Field(default=None, description="Ramsey detuning in Hz")

    # Pulse parameters
    ctrlpulse_pi_duration_ns: int = Field(description="Control pulse pi duration in ns", gt=0)
    ctrlpulse_shape: str = Field(description="Control pulse shape")
    ctrlpulse_pi_amplitude: float | None = Field(
        default=None, description="Control pulse pi amplitude"
    )
    readpulse_duration_ns: int = Field(description="Readout pulse duration in ns", gt=0)
    readpulse_shape: str = Field(description="Readout pulse shape")
    readpulse_amplitude: float = Field(description="Readout pulse amplitude")
    reset_time_us: float = Field(description="Reset time in microseconds")

    # Qubit and resonator properties
    qubit_frequency: float = Field(description="Qubit frequency in Hz", gt=0)
    resonator_frequency: float = Field(description="Resonator frequency in Hz", gt=0)
    resonator_qtotal: float = Field(description="Resonator total Q factor")
    qubit_ej_ghz: float | None = Field(default=None, description="Qubit EJ in GHz")
    qubit_ec_ghz: float | None = Field(default=None, description="Qubit EC in GHz")
    qubit_el_ghz: float | None = Field(default=None, description="Qubit EL in GHz")
    anharmonicity_mhz: float | None = Field(default=None, description="Anharmonicity in MHz")

    # Material properties
    sample_origin_cleanroom: str = Field(description="Sample origin cleanroom")
    material_substrate: str = Field(description="Substrate material")
    material_superconductor: str = Field(description="Superconductor material")
    material_jj: str = Field(description="Josephson junction material")
    fabrication_date: str | None = Field(default=None, description="Fabrication date")

    # Amplifier and temperature
    parametric_amplifier_used: str = Field(description="Parametric amplifier identifier")
    parametric_amplifier_gain_db: float | None = Field(
        default=None, description="Amplifier gain in dB"
    )
    base_temperature_mk: float | None = Field(default=None, description="Base temperature in mK")
    chip_package: str | None = Field(default=None, description="Chip package type")

    # Wiring configuration
    control_line_number: str = Field(description="Control line number")
    control_line_config: str = Field(description="Control line configuration")
    readout_in_line_number: str = Field(description="Readout input line number")
    readout_in_line_config: str = Field(description="Readout input line configuration")
    readout_out_line_number: str = Field(description="Readout output line number")
    readout_out_line_config: str = Field(description="Readout output line configuration")
    fridge_wiring_reference: str = Field(description="Fridge wiring reference")

    # References and history
    data_reference: str = Field(description="Data reference")
    tuneup_measurement_reference: str = Field(description="Tuneup measurement reference")
    sample_history_reference: str | None = Field(
        default=None, description="Sample history reference"
    )

    # Sweep parameters (optional)
    part_of_sweep: bool | None = Field(default=None, description="Whether part of a sweep")
    sweep_variable: str | None = Field(default=None, description="Sweep variable name")
    sweep_variable_value: float | None = Field(default=None, description="Sweep variable value")
    sweep_variable_unit: str | None = Field(default=None, description="Sweep variable unit")
    sweep_variable_list: str | None = Field(default=None, description="Sweep variable list (JSON)")
    sweep_reference_next: str | None = Field(default=None, description="Next sweep reference")
    sweep_reference_previous: str | None = Field(
        default=None, description="Previous sweep reference"
    )


class MeasurementPoint(BaseModel):
    """A single quantum measurement data point for InfluxDB."""

    measurement: str = Field(description="InfluxDB measurement name")
    tags: MeasurementTags = Field(description="Measurement tags")
    fields: T1MeasurementFields | T2EchoMeasurementFields | T2RamseyMeasurementFields = Field(
        description="Measurement fields"
    )
    timestamp: datetime = Field(description="Measurement timestamp")
    tuid: str | None = Field(default=None, description="Original TUID from analysis")

    @field_validator("measurement")
    @classmethod
    def validate_measurement_name(cls, v: str) -> str:
        """Validate InfluxDB measurement name format."""
        if not v.replace("_", "").isalnum():
            raise ValueError("Measurement name must be alphanumeric with underscores")
        return v

    @model_validator(mode="after")
    def validate_measurement_type_matches_fields(self) -> MeasurementPoint:
        """Validate that measurement type tag matches the fields model type."""
        measurement_type = self.tags.measurement_type
        actual_fields_type = type(self.fields).__name__

        expected_fields_type_mapping = {
            MeasurementType.T1: "T1MeasurementFields",
            MeasurementType.T2_ECHO: "T2EchoMeasurementFields",
            MeasurementType.T2_RAMSEY: "T2RamseyMeasurementFields",
        }

        expected_fields_type = expected_fields_type_mapping.get(measurement_type)
        if expected_fields_type != actual_fields_type:
            raise ValueError(
                f"Measurement type mismatch: {measurement_type} requires "
                f"{expected_fields_type}, got {actual_fields_type}"
            )

        return self

    def to_influx_dict(self) -> dict[str, Any]:
        """Convert to InfluxDB line protocol dictionary."""
        return {
            "measurement": self.measurement,
            "tags": self.tags.model_dump(exclude_none=True),
            "fields": self.fields.model_dump(exclude_none=True),
            "time": self.timestamp,
        }


class AnalysisMetadataTags(BaseModel):
    """Tags associated with analysis metadata points."""

    measurement_type: str = Field(description="Measurement grouping from metadata")
    experiment_id: str = Field(description="Experiment unique identifier")
    device_name: str = Field(description="Quantum device name")
    qubit_name: str = Field(description="Qubit name")

    @field_validator("experiment_id", "device_name", "qubit_name")
    @classmethod
    def validate_identifiers(cls, v: str) -> str:
        """Validate identifier format (no spaces, special chars)."""
        if not v.replace("_", "").replace("-", "").isalnum():
            raise ValueError("Identifiers must be alphanumeric with _ or - only")
        return v


class AnalysisMetadataFields(BaseModel):
    """Flexible metadata fields for analysis points."""

    model_config = ConfigDict(extra="allow")

    measurement_time: str = Field(description="Timestamp recorded in metadata (ISO string)")


class AnalysisMetadataPoint(BaseModel):
    """Metadata snapshot associated with an analysis run."""

    measurement: str = Field(description="InfluxDB measurement name")
    tags: AnalysisMetadataTags = Field(description="Metadata tags")
    fields: AnalysisMetadataFields = Field(description="Metadata fields")
    timestamp: datetime = Field(description="Metadata timestamp")
    tuid: str | None = Field(default=None, description="Original TUID from analysis")

    def to_influx_dict(self) -> dict[str, Any]:
        """Convert to InfluxDB line protocol dictionary."""
        return {
            "measurement": self.measurement,
            "tags": self.tags.model_dump(exclude_none=True),
            "fields": self.fields.model_dump(exclude_none=True),
            "time": self.timestamp,
        }



"""Utility functions."""

from .utils import (
    check_and_create_dir,
    read_json,
    serialize_for_json,
    setup_logging,
    write_json,
)

__all__ = [
    "check_and_create_dir",
    "read_json",
    "serialize_for_json",
    "setup_logging",
    "write_json",
]



"""Utility functions for measurement analysis."""

from __future__ import annotations

import pathlib
from datetime import datetime
from functools import lru_cache
from typing import Any

from metsuperq.analysis.constants import MAX_PROJECT_ROOT_DEPTH, SECONDS_TO_MICROSECONDS
from metsuperq.utils import read_json, setup_logging, write_json

logger = setup_logging(__name__)


@lru_cache(maxsize=1)
def get_project_root() -> pathlib.Path:
    """Find project root directory.

    Searches upward from current location for project markers like:
    - pyproject.toml
    - setup.py
    - .git directory
    - metsuperq package directory

    Returns
    -------
    pathlib.Path
        Absolute path to project root
    """
    # Start from the module's location
    current_path = pathlib.Path(__file__).resolve().parent

    # Prefer explicit repository markers
    repo_markers = ["pyproject.toml", ".git", "setup.py"]

    # Traverse upwards with a sane max depth to avoid infinite loops under mocks
    for _ in range(MAX_PROJECT_ROOT_DEPTH):
        # Check for explicit repo markers first using any() for early exit
        try:
            found_marker = next((m for m in repo_markers if (current_path / m).exists()), None)
            if found_marker:
                logger.debug("Auto-detected project root via '%s': %s", found_marker, current_path)
                return current_path
        except OSError:
            logger.debug("Failed to check for repository markers at %s", current_path)

        # If we are at the package directory (e.g., 'metsuperq'), prefer its parent as project root
        try:
            if getattr(current_path, "name", None) == "metsuperq":
                parent = current_path.parent
                # If parent has typical repo markers, treat it as the project root
                if (parent / "pyproject.toml").exists() or (parent / ".git").exists():
                    logger.debug("Detected project root at package parent: %s", parent)
                    return parent
        except OSError:
            logger.debug("Failed to check for repository markers at %s", current_path)

        # Move up one directory
        try:
            current_path = current_path.parent
        except (OSError, AttributeError):
            logger.debug("Failed to move up one directory at %s", current_path)
            break

    # Fallback: use current working directory
    fallback_path = pathlib.Path.cwd()
    logger.warning("Could not find project root via markers, using fallback: %s", fallback_path)
    return fallback_path


# Type aliases
AnalysisResult = dict[str, Any]


def save_analysis_results(analysis_result: AnalysisResult, output_path: pathlib.Path) -> None:
    """Save analysis results to JSON file.

    Parameters
    ----------
    analysis_result : AnalysisResult
        Analysis results to save
    output_path : pathlib.Path
        Path for output JSON file
    """
    logger.debug("Saving analysis results to: %s", output_path)

    # Add analysis timestamp
    result_with_timestamp = analysis_result.copy()
    result_with_timestamp["analysis_timestamp"] = datetime.now().isoformat()

    try:
        write_json(result_with_timestamp, output_path)
        logger.info("Analysis results saved to: %s", output_path)
    except OSError as e:
        logger.error("Failed to save analysis results: %s", e, exc_info=True)
        raise


def load_analysis_results(input_path: pathlib.Path) -> AnalysisResult:
    """Load analysis results from JSON file.

    Parameters
    ----------
    input_path : pathlib.Path
        Path to JSON file containing analysis results

    Returns
    -------
    AnalysisResult
        Loaded analysis results

    Raises
    ------
    FileNotFoundError
        If the results file doesn't exist
    """
    logger.debug("Loading analysis results from: %s", input_path)

    try:
        results = read_json(input_path)
        logger.info("Analysis results loaded from: %s", input_path)
        return results
    except (OSError, ValueError) as e:
        logger.error("Failed to load analysis results: %s", e, exc_info=True)
        raise


def create_summary_report(analysis_result: AnalysisResult) -> str:  # noqa: C901, PLR0912, PLR0915
    """Create a text summary report of analysis results.

    Parameters
    ----------
    analysis_result : AnalysisResult
        Analysis results to summarize

    Returns
    -------
    str
        Formatted summary report
    """
    logger.debug("Creating summary report")

    report_lines = []
    report_lines.append("=== Measurement Analysis Report ===")
    report_lines.append("")

    # Basic info
    filepath = analysis_result.get("filepath", "Unknown")
    timestamp = analysis_result.get("analysis_timestamp", "Unknown")
    report_lines.append(f"Data file: {filepath}")
    report_lines.append(f"Analysis timestamp: {timestamp}")
    report_lines.append("")

    # Fit results
    fits = analysis_result.get("fits", {})
    if fits:
        report_lines.append("=== Fit Results ===")
        for meas_type, fit_result in fits.items():
            if fit_result.get("fit_success", False):
                decay_time = fit_result.get("decay_time", 0)
                decay_stderr = fit_result.get("decay_time_stderr", 0)
                amplitude = fit_result.get("amplitude", 0)
                r_squared = fit_result.get("r_squared", 0)

                report_lines.append(f"{meas_type.upper()}:")
                # Convert to microseconds for better readability
                decay_time_us = decay_time * SECONDS_TO_MICROSECONDS
                decay_stderr_us = decay_stderr * SECONDS_TO_MICROSECONDS
                report_lines.append(f"  Decay time: {decay_time_us:.1f} ± {decay_stderr_us:.1f} μs")
                report_lines.append(f"  Amplitude: {amplitude:.3f}")
                report_lines.append(f"  R²: {r_squared:.3f}")
            else:
                report_lines.append(f"{meas_type.upper()}: Fit failed")
            report_lines.append("")

    # Derived quantities
    derived = analysis_result.get("derived_quantities", {})
    if derived:
        report_lines.append("=== Derived Quantities ===")
        for key, value in derived.items():
            if not key.endswith("_stderr") and isinstance(value, (int, float)):
                stderr_key = f"{key}_stderr"
                stderr = derived.get(stderr_key, 0)
                clean_key = key.replace("_", " ").title()
                # Convert to microseconds for better readability
                value_us = value * SECONDS_TO_MICROSECONDS
                stderr_us = stderr * SECONDS_TO_MICROSECONDS
                report_lines.append(f"{clean_key}: {value_us:.1f} ± {stderr_us:.1f} μs")

    # Physics validation results
    physics_validation = analysis_result.get("physics_validation", {})
    if physics_validation:
        report_lines.append("")
        report_lines.append("=== Physics Validation ===")
        is_valid = physics_validation.get("is_valid", True)
        violations = physics_validation.get("violations", [])
        warnings = physics_validation.get("warnings", [])

        if is_valid:
            report_lines.append("Status: ✓ PASSED - All coherence times are physically valid")
        else:
            report_lines.append("Status: ✗ FAILED - Physics constraints violated!")

        if violations:
            report_lines.append("")
            report_lines.append("Violations:")
            for violation in violations:
                report_lines.append(f"  • {violation}")

        if warnings:
            report_lines.append("")
            report_lines.append("Warnings:")
            for warning in warnings:
                report_lines.append(f"  • {warning}")

    report = "\n".join(report_lines)
    logger.debug("Summary report created with %d lines", len(report_lines))
    return report


def save_summary_report(analysis_result: AnalysisResult, output_path: pathlib.Path) -> None:
    """Save a text summary report to file.

    Parameters
    ----------
    analysis_result : AnalysisResult
        Analysis results to summarize
    output_path : pathlib.Path
        Path for output text file
    """
    logger.debug("Saving summary report to: %s", output_path)

    report = create_summary_report(analysis_result)

    try:
        output_path.parent.mkdir(parents=True, exist_ok=True)
        with open(output_path, "w", encoding="utf-8") as f:
            f.write(report)
        logger.info("Summary report saved to: %s", output_path)
    except OSError as e:
        logger.error("Failed to save summary report: %s", e, exc_info=True)
        raise



"""Configuration management for InfluxDB connection."""

from __future__ import annotations

from typing import Any, Self

from pydantic import Field, ValidationError
from pydantic_settings import BaseSettings

from metsuperq.utils import setup_logging

logger = setup_logging(__name__)


class InfluxConfigError(Exception):
    """Exception raised for InfluxDB configuration errors.

    This exception provides user-friendly error messages for common
    configuration issues like missing environment variables.
    """

    pass


class InfluxConfig(BaseSettings):
    """Configuration for InfluxDB v2 connection.

    Attributes
    ----------
    host : str
        InfluxDB host (required)
        InfluxDB port (required)
    token : str
        Authentication token (required)
    org : str
        Organization name (required)
    bucket : str
        Bucket name for measurements (required)
    timeout : int
        Request timeout in milliseconds (optional, default: 30000)
    verify_ssl : bool
        Whether to verify SSL certificates (optional, default: True)
    """

    host: str = Field(description="InfluxDB host")
    port: int = Field(description="InfluxDB port")
    token: str = Field(description="Authentication token")
    org: str = Field(description="Organization name")
    bucket: str = Field(description="Bucket name for measurements")
    timeout: int = Field(default=30000, description="Request timeout in ms")
    verify_ssl: bool = Field(default=True, description="Verify SSL certificates")

    model_config = {
        "env_file": ".env",
        "env_file_encoding": "utf-8",
        "case_sensitive": False,
        "env_prefix": "INFLUXDB_",
    }

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    @property
    def url(self) -> str:
        """Return InfluxDB URL."""
        protocol = "https" if self.verify_ssl else "http"
        return f"{protocol}://{self.host}:{self.port}"

    @classmethod
    def from_env(cls) -> Self:
        """Create configuration from environment variables and .env file.

        Automatically loads from .env file if present, then environment variables.

        Raises
        ------
        InfluxConfigError
            If required environment variables are missing
        """
        logger.debug("Loading InfluxDB configuration from environment and .env file")

        try:
            return cls()
        except ValidationError as e:
            # Use list comprehension to extract missing field names
            missing_fields = [error["loc"][0] for error in e.errors() if error["type"] == "missing"]

            if missing_fields:
                field_mapping = {
                    "token": "INFLUXDB_TOKEN",
                    "org": "INFLUXDB_ORG",
                    "bucket": "INFLUXDB_BUCKET",
                    "host": "INFLUXDB_HOST",
                    "port": "INFLUXDB_PORT",
                    "timeout": "INFLUXDB_TIMEOUT",
                    "verify_ssl": "INFLUXDB_VERIFY_SSL",
                }

                # Map field names to environment variable names
                env_vars = [field_mapping[f] for f in missing_fields if f in field_mapping]

                logger.error("Missing required InfluxDB configuration: %s", missing_fields)
                raise InfluxConfigError(
                    "Missing required InfluxDB configuration.\n\n"
                    "Required environment variables (or .env file entries):\n"
                    + "\n".join(f"  • {var}" for var in env_vars)
                    + "\n\n"
                    "Set these in your .env file:\n"
                    "  INFLUXDB_URL='your-url-here:your-port-here'\n"
                    "  INFLUXDB_TOKEN=your-token-here\n"
                    "  INFLUXDB_ORG=your-org-name\n"
                    "  INFLUXDB_BUCKET=your-bucket-name\n\n"
                    "Or as environment variables:\n"
                    "  export INFLUXDB_URL='your-url-here:your-port-here'\n"
                    "  export INFLUXDB_TOKEN='your-token-here'\n"
                    "  export INFLUXDB_ORG='your-org-name'\n"
                    "  export INFLUXDB_BUCKET='your-bucket-name'\n\n"
                    "Optional variables:\n"
                    "  INFLUXDB_TIMEOUT=30000\n"
                    "  INFLUXDB_VERIFY_SSL=true\n"
                ) from e

            # Re-raise other validation errors
            raise

    def mask_sensitive(self) -> dict[str, Any]:
        """Return config dict with sensitive data masked."""
        config_dict = self.model_dump()
        if config_dict.get("token"):
            config_dict["token"] = f"***{config_dict['token'][-4:]}"
        return config_dict



"""Utility functions."""

from __future__ import annotations

import json
import logging
import os
from pathlib import Path
from typing import Any

import numpy as np


def setup_logging(name: str, level: str | None = None) -> logging.Logger:
    """Set up logging with enhanced configuration for quantum measurements.

    Parameters
    ----------
    name : str
        Name of the logger (typically __name__)
    level : str | None, optional
        Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL).
        If None, uses LOG_LEVEL environment variable or defaults to DEBUG.

    Returns
    -------
    logging.Logger
        Configured logger instance with enhanced formatting
    """
    logger = logging.getLogger(name)

    # Only configure if not already configured to avoid duplicate handlers
    if not logger.handlers:
        # Determine log level from parameter or environment
        if level is None:
            level = os.getenv("LOG_LEVEL", "DEBUG").upper()

        # Convert string level to logging level constant
        level_constant = getattr(logging, level.upper(), logging.INFO)
        logger.setLevel(level_constant)

        # Create handler with enhanced formatting
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s | %(name)s | %(levelname)s | %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        handler.setFormatter(formatter)
        handler.setLevel(level_constant)
        logger.addHandler(handler)

        # Prevent propagation to avoid duplicate logs
        logger.propagate = False

    return logger


def ensure_directory_exists(directory: Path | str) -> None:
    """Ensure a directory exists, creating it and parents if necessary.

    Parameters
    ----------
    directory : Path | str
        Directory path to ensure exists
    """
    Path(directory).mkdir(parents=True, exist_ok=True)


# Backward compatibility alias
check_and_create_dir = ensure_directory_exists


def serialize_for_json(obj: Any) -> Any:
    """Convert numpy types and other objects for JSON serialization.

    Parameters
    ----------
    obj : Any
        Object to convert for JSON serialization

    Returns
    -------
    Any
        JSON-serializable object

    Raises
    ------
    TypeError
        If the object cannot be serialized to JSON
    """
    try:
        # Handle different types with consolidated logic
        conversion_map = {
            np.integer: int,
            np.floating: float,
        }

        for obj_type, converter in conversion_map.items():
            if isinstance(obj, obj_type):
                return converter(obj)

        # Handle special cases
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, dict):
            return {k: serialize_for_json(v) for k, v in obj.items()}
        elif isinstance(obj, (list, tuple)):
            return [serialize_for_json(item) for item in obj]
        elif isinstance(obj, (str, int, float, bool, type(None))):
            return obj
        else:
            # Try to serialize the object to JSON to see if it's serializable
            json.dumps(obj)
            return obj
    except (TypeError, OverflowError) as e:
        raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable") from e


def read_json(file_path: Path | str) -> Any:
    """Read a JSON file and return its contents.

    Parameters
    ----------
    file_path : Path | str
        Path to JSON file

    Returns
    -------
    Any
        Parsed JSON data

    Raises
    ------
    FileNotFoundError
        If file doesn't exist
    json.JSONDecodeError
        If file contains invalid JSON
    """
    try:
        with open(file_path) as f:
            return json.load(f)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"JSON file not found: {file_path}") from e
    except json.JSONDecodeError as e:
        raise json.JSONDecodeError(f"Invalid JSON in file: {file_path}", e.doc, e.pos) from e


def write_json(data: Any, file_path: Path | str, indent: int = 2) -> None:
    """Write data to a JSON file with proper serialization.

    Parameters
    ----------
    data : Any
        Data to write to JSON
    file_path : Path | str
        Output file path
    indent : int, optional
        JSON indentation, by default 2

    Raises
    ------
    TypeError
        If data cannot be serialized to JSON
    """
    # Ensure directory exists
    Path(file_path).parent.mkdir(parents=True, exist_ok=True)

    # Serialize data
    serialized_data = serialize_for_json(data)

    try:
        with open(file_path, "w") as f:
            json.dump(serialized_data, f, indent=indent)
    except TypeError as e:
        raise TypeError(f"Cannot serialize data to JSON: {e}") from e



"""Simple synchronous InfluxDB v2 client wrapper."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from influxdb_client.client.query_api import QueryApi
    from influxdb_client.client.write_api import WriteApi

from influxdb_client.client.exceptions import InfluxDBError
from influxdb_client.client.influxdb_client import InfluxDBClient as _InfluxDBClient
from influxdb_client.client.write_api import SYNCHRONOUS

from metsuperq.database.config import InfluxConfig
from metsuperq.database.exceptions import InfluxDBConnectionError
from metsuperq.utils import setup_logging

logger = setup_logging(__name__)


class InfluxDBClient:
    """Synchronous InfluxDB v2 client wrapper."""

    def __init__(self, config: InfluxConfig) -> None:
        """Initialize client with configuration.

        Parameters
        ----------
        config : InfluxConfig
            InfluxDB connection configuration
        """
        self.config = config
        self._client: _InfluxDBClient | None = None
        self._write_api: WriteApi | None = None
        self._query_api: QueryApi | None = None
        logger.info("Initialized InfluxDBClient for %s bucket=%s", config.url, config.bucket)

    def connect(self) -> None:
        """Establish connection to InfluxDB."""
        try:
            logger.debug("Connecting to InfluxDB at %s", self.config.url)

            self._client = _InfluxDBClient(
                url=self.config.url,
                token=self.config.token,
                org=self.config.org,
                timeout=self.config.timeout,
                verify_ssl=self.config.verify_ssl,
            )

            # Test connection
            if self._client is not None:
                self._client.ping()

                # Initialize APIs with SYNCHRONOUS writes
                self._write_api = self._client.write_api(write_options=SYNCHRONOUS)
                self._query_api = self._client.query_api()

            logger.info("Successfully connected to InfluxDB")

        except InfluxDBError as e:
            logger.error("Failed to connect to InfluxDB: %s", e, exc_info=True)
            raise InfluxDBConnectionError(
                f"Failed to connect to InfluxDB at {self.config.url}", details={"error": str(e)}
            ) from e

    def disconnect(self) -> None:
        """Close connection to InfluxDB."""
        if self._client:
            try:
                logger.debug("Disconnecting from InfluxDB")
                self._client.close()
                self._client = None
                self._write_api = None
                self._query_api = None
                logger.info("Disconnected from InfluxDB")
            except Exception as e:
                logger.warning("Error during disconnect: %s", e)

    def __enter__(self) -> InfluxDBClient:
        """Context manager entry."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Context manager exit."""
        self.disconnect()

    @property
    def is_connected(self) -> bool:
        """Check if client is connected."""
        if self._client is None:
            return False
        try:
            return self._client.ping()
        except Exception:
            return False

    def get_write_api(self) -> WriteApi:
        """Get write API instance."""
        if not self._write_api:
            raise InfluxDBConnectionError("Client not connected")
        return self._write_api

    def get_query_api(self) -> QueryApi:
        """Get query API instance."""
        if not self._query_api:
            raise InfluxDBConnectionError("Client not connected")
        return self._query_api

    def health_check(self) -> dict[str, Any]:
        """Perform health check and return status."""
        try:
            if not self._client:
                return {"status": "disconnected", "healthy": False}

            # Ping check
            ping_result = self._client.ping()

            # Bucket existence check
            buckets_api = self._client.buckets_api()
            bucket_exists = buckets_api.find_bucket_by_name(self.config.bucket)

            return {
                "status": "connected",
                "healthy": ping_result and bucket_exists is not None,
                "bucket_exists": bucket_exists is not None,
                "config": self.config.mask_sensitive(),
            }
        except Exception as e:
            logger.error("Health check failed: %s", e, exc_info=True)
            return {
                "status": "error",
                "healthy": False,
                "error": str(e),
            }

    def close(self) -> None:
        """Close the client connection (alias for disconnect)."""
        self.disconnect()



"""Data handling utilities for MetSuperQ.

Exposes helpers for interacting with HDF5 files.
"""

from . import npl_converter
from .hdf5_data_manager import HDF5DataManager

__all__ = [
    "HDF5DataManager",
    "npl_converter",
]



"""Monitoring and metrics collection for MetSuperQ."""

from __future__ import annotations

from metsuperq.monitoring.health import HealthChecker, SystemHealth

__all__ = [
    "HealthChecker",
    "SystemHealth",
]



"""Health checking and system monitoring for MetSuperQ."""

from __future__ import annotations

import os
import tempfile
import time
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any

import h5py
import lmfit
import numpy as np
import psutil
from pydantic import BaseModel, Field

from metsuperq.analysis.base_analysis import BaseAnalyzer
from metsuperq.database.client import InfluxDBClient
from metsuperq.database.config import InfluxConfig
from metsuperq.database.exceptions import InfluxDBConnectionError
from metsuperq.utils import setup_logging

logger = setup_logging(__name__)

# Health check thresholds
LOW_DISK_SPACE_GB = 1.0  # Disk space below this triggers degraded status
CRITICAL_MEMORY_PERCENT = 90  # Memory usage above this is critical
HIGH_MEMORY_PERCENT = 80  # Memory usage above this triggers warning

# Unit conversion constants
BYTES_PER_GB = 1024**3  # Bytes in a gigabyte
BYTES_PER_MB = 1024**2  # Bytes in a megabyte
MS_PER_SECOND = 1000  # Milliseconds per second


class HealthStatus(str, Enum):
    """Health check status levels."""

    HEALTHY = "healthy"
    DEGRADED = "degraded"
    UNHEALTHY = "unhealthy"
    UNKNOWN = "unknown"


class ComponentType(str, Enum):
    """Types of system components to monitor."""

    DATABASE = "database"
    ANALYSIS_ENGINE = "analysis_engine"
    FILE_SYSTEM = "file_system"
    MEMORY = "memory"
    NETWORK = "network"


@dataclass
class HealthCheck:
    """Individual health check result."""

    component: ComponentType
    status: HealthStatus
    message: str
    response_time_ms: float
    timestamp: float
    details: dict[str, Any] | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert health check to dictionary."""
        return {
            "component": self.component.value,
            "status": self.status.value,
            "message": self.message,
            "response_time_ms": self.response_time_ms,
            "timestamp": self.timestamp,
            "details": self.details or {},
        }


class SystemHealth(BaseModel):
    """Overall system health status."""

    overall_status: HealthStatus = Field(description="Overall system health")
    checks: list[HealthCheck] = Field(default_factory=list, description="Individual health checks")
    timestamp: float = Field(default_factory=time.time, description="Health check timestamp")
    uptime_seconds: float = Field(ge=0, description="System uptime")
    message: str = Field(description="Health check message")

    def add_check(self, check: HealthCheck) -> None:
        """Add a health check result and recalculate overall status."""
        self.checks.append(check)
        self._recalculate_overall_status_from_checks()

    def _recalculate_overall_status_from_checks(self) -> None:
        """Recalculate overall status based on all individual check results."""
        if not self.checks:
            self.overall_status = HealthStatus.UNKNOWN
            return

        statuses = [check.status for check in self.checks]

        if all(status == HealthStatus.HEALTHY for status in statuses):
            self.overall_status = HealthStatus.HEALTHY
        elif any(status == HealthStatus.UNHEALTHY for status in statuses):
            self.overall_status = HealthStatus.UNHEALTHY
        elif any(status == HealthStatus.DEGRADED for status in statuses):
            self.overall_status = HealthStatus.DEGRADED
        else:
            self.overall_status = HealthStatus.UNKNOWN

    def get_unhealthy_components(self) -> list[HealthCheck]:
        """Get list of unhealthy components."""
        return [
            check
            for check in self.checks
            if check.status in (HealthStatus.UNHEALTHY, HealthStatus.DEGRADED)
        ]

    def to_dict(self) -> dict[str, Any]:
        """Convert system health to dictionary."""
        return {
            "overall_status": self.overall_status.value,
            "checks": [check.to_dict() for check in self.checks],
            "timestamp": self.timestamp,
            "uptime_seconds": self.uptime_seconds,
            "message": self.message,
        }


class HealthChecker:
    """Comprehensive health checking for MetSuperQ components."""

    def __init__(self, influx_config: InfluxConfig | None = None) -> None:
        """Initialize health checker.

        Parameters
        ----------
        influx_config : InfluxConfig | None
            InfluxDB configuration for database health checks
        """
        self._influx_config = influx_config
        self._start_time = time.time()

        logger.info("Initialized HealthChecker")

    def check_database_health(self) -> HealthCheck:
        """Check InfluxDB database health.

        Returns
        -------
        HealthCheck
            Database health check result
        """
        start_time = time.time()

        if not self._influx_config:
            return HealthCheck(
                component=ComponentType.DATABASE,
                status=HealthStatus.UNKNOWN,
                message="No database configuration provided",
                response_time_ms=0.0,
                timestamp=time.time(),
            )

        try:
            # Use context manager to ensure connection is always closed
            with InfluxDBClient(self._influx_config) as client:
                # All database operations happen within connection scope
                health_info = client.health_check()

                response_time = (time.time() - start_time) * MS_PER_SECOND

                if health_info and health_info.get("status") == "pass":
                    return HealthCheck(
                        component=ComponentType.DATABASE,
                        status=HealthStatus.HEALTHY,
                        message="Database connection successful",
                        response_time_ms=response_time,
                        timestamp=time.time(),
                        details=health_info,
                    )
                else:
                    return HealthCheck(
                        component=ComponentType.DATABASE,
                        status=HealthStatus.DEGRADED,
                        message="Database responding but not fully healthy",
                        response_time_ms=response_time,
                        timestamp=time.time(),
                        details=health_info,
                    )

        except InfluxDBConnectionError as e:
            response_time = (time.time() - start_time) * 1000
            return HealthCheck(
                component=ComponentType.DATABASE,
                status=HealthStatus.UNHEALTHY,
                message=f"Database connection failed: {e}",
                response_time_ms=response_time,
                timestamp=time.time(),
                details={"error": str(e)},
            )
        except Exception as e:
            response_time = (time.time() - start_time) * 1000
            return HealthCheck(
                component=ComponentType.DATABASE,
                status=HealthStatus.UNHEALTHY,
                message=f"Unexpected database error: {e}",
                response_time_ms=response_time,
                timestamp=time.time(),
                details={"error": str(e)},
            )

    def check_analysis_engine_health(self) -> HealthCheck:
        """Check analysis engine health.

        Returns
        -------
        HealthCheck
            Analysis engine health check result
        """
        start_time = time.time()

        try:
            # Basic functionality test
            _ = BaseAnalyzer()

            response_time = (time.time() - start_time) * 1000

            return HealthCheck(
                component=ComponentType.ANALYSIS_ENGINE,
                status=HealthStatus.HEALTHY,
                message="Analysis engine initialized successfully",
                response_time_ms=response_time,
                timestamp=time.time(),
                details={
                    "numpy_version": np.__version__,
                    "lmfit_version": lmfit.__version__,
                    "h5py_version": h5py.__version__,
                },
            )

        except ImportError as e:
            response_time = (time.time() - start_time) * 1000
            return HealthCheck(
                component=ComponentType.ANALYSIS_ENGINE,
                status=HealthStatus.UNHEALTHY,
                message=f"Missing dependency: {e}",
                response_time_ms=response_time,
                timestamp=time.time(),
                details={"error": str(e)},
            )
        except Exception as e:
            response_time = (time.time() - start_time) * 1000
            return HealthCheck(
                component=ComponentType.ANALYSIS_ENGINE,
                status=HealthStatus.DEGRADED,
                message=f"Analysis engine issue: {e}",
                response_time_ms=response_time,
                timestamp=time.time(),
                details={"error": str(e)},
            )

    def check_file_system_health(self) -> HealthCheck:
        """Check file system health for data directories.

        Returns
        -------
        HealthCheck
            File system health check result
        """
        start_time = time.time()

        try:
            # Test write/read in temp directory
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
                test_file = Path(f.name)
                f.write("health_check_test")

            # Verify file was written and can be read
            content = test_file.read_text()
            test_file.unlink()  # Clean up

            if content != "health_check_test":
                raise ValueError("File system read/write test failed")

            # Check available disk space
            stat = os.statvfs(tempfile.gettempdir())
            free_space_gb = (stat.f_frsize * stat.f_bavail) / BYTES_PER_GB

            response_time = (time.time() - start_time) * MS_PER_SECOND

            if free_space_gb < LOW_DISK_SPACE_GB:
                status = HealthStatus.DEGRADED
                message = f"Low disk space: {free_space_gb:.1f}GB free"
            else:
                status = HealthStatus.HEALTHY
                message = f"File system healthy: {free_space_gb:.1f}GB free"

            return HealthCheck(
                component=ComponentType.FILE_SYSTEM,
                status=status,
                message=message,
                response_time_ms=response_time,
                timestamp=time.time(),
                details={"free_space_gb": free_space_gb},
            )

        except Exception as e:
            response_time = (time.time() - start_time) * MS_PER_SECOND
            return HealthCheck(
                component=ComponentType.FILE_SYSTEM,
                status=HealthStatus.UNHEALTHY,
                message=f"File system error: {e}",
                response_time_ms=response_time,
                timestamp=time.time(),
                details={"error": str(e)},
            )

    def check_memory_health(self) -> HealthCheck:
        """Check memory usage health.

        Returns
        -------
        HealthCheck
            Memory health check result
        """
        start_time = time.time()

        try:
            if psutil is None:
                raise ImportError("psutil not available")

            memory = psutil.virtual_memory()
            process = psutil.Process()
            process_memory_mb = process.memory_info().rss / BYTES_PER_MB

            response_time = (time.time() - start_time) * MS_PER_SECOND

            # Determine status based on memory usage
            if memory.percent > CRITICAL_MEMORY_PERCENT:
                status = HealthStatus.UNHEALTHY
                message = f"Critical memory usage: {memory.percent:.1f}%"
            elif memory.percent > HIGH_MEMORY_PERCENT:
                status = HealthStatus.DEGRADED
                message = f"High memory usage: {memory.percent:.1f}%"
            else:
                status = HealthStatus.HEALTHY
                message = f"Memory usage normal: {memory.percent:.1f}%"

            return HealthCheck(
                component=ComponentType.MEMORY,
                status=status,
                message=message,
                response_time_ms=response_time,
                timestamp=time.time(),
                details={
                    "system_memory_percent": memory.percent,
                    "system_memory_available_gb": memory.available / BYTES_PER_GB,
                    "process_memory_mb": process_memory_mb,
                },
            )

        except ImportError:
            response_time = (time.time() - start_time) * MS_PER_SECOND
            return HealthCheck(
                component=ComponentType.MEMORY,
                status=HealthStatus.UNKNOWN,
                message="psutil not available for memory monitoring",
                response_time_ms=response_time,
                timestamp=time.time(),
            )
        except Exception as e:
            response_time = (time.time() - start_time) * MS_PER_SECOND
            return HealthCheck(
                component=ComponentType.MEMORY,
                status=HealthStatus.DEGRADED,
                message=f"Memory check error: {e}",
                response_time_ms=response_time,
                timestamp=time.time(),
                details={"error": str(e)},
            )

    def perform_full_health_check(self) -> SystemHealth:
        """Perform comprehensive health check of all components.

        Returns
        -------
        SystemHealth
            Complete system health status
        """
        logger.info("Starting full system health check")

        health = SystemHealth(
            uptime_seconds=time.time() - self._start_time,
            timestamp=time.time(),
            overall_status=HealthStatus.HEALTHY,
            message="System health check completed",
        )

        # Perform all health checks
        checks = [
            self.check_analysis_engine_health(),
            self.check_file_system_health(),
            self.check_memory_health(),
        ]

        # Add database check if configuration is available
        if self._influx_config:
            checks.append(self.check_database_health())

        # Add all checks to system health
        for check in checks:
            health.add_check(check)

        logger.info(
            "Health check completed",
            extra={
                "overall_status": health.overall_status.value,
                "unhealthy_components": len(health.get_unhealthy_components()),
                "total_checks": len(health.checks),
            },
        )

        return health

    def get_health_summary(self) -> dict[str, Any]:
        """Get a quick health summary.

        Returns
        -------
        dict[str, Any]
            Health summary dictionary
        """
        health = self.perform_full_health_check()
        return {
            "status": health.overall_status.value,
            "timestamp": health.timestamp,
            "uptime_seconds": health.uptime_seconds,
            "components_checked": len(health.checks),
            "unhealthy_components": len(health.get_unhealthy_components()),
            "details": health.to_dict(),
        }



"""Constants for the base measurement analyzer."""

from __future__ import annotations

# Measurement types
MEASUREMENT_TYPES = ["t1", "t2e", "t2r"]

# Plot configuration
PLOT_CONFIG = {
    "figsize_per_plot": 5,
    "height": 4,
    "dpi": 300,
    "alpha": 0.7,
    "grid_alpha": 0.3,
}

# File patterns
HDF5_PATTERN = "**/*.hdf5"
DEFAULT_MEASUREMENT_TYPE = "coherence_dataset"

# Fit parameters
DEFAULT_FIT_POINTS = 100

# Data structure constants
EXPECTED_DATASET_ROWS = 2  # Expected number of rows in dataset (time, IQ data)
MIN_POINTS_FOR_FFT = 3  # Minimum points needed for FFT frequency estimation

# Function complexity limits
MAX_RETURN_STATEMENTS = 6  # Maximum return statements allowed in functions
MAX_FUNCTION_ARGUMENTS = 5  # Maximum arguments allowed in function definition
MAX_FUNCTION_STATEMENTS = 50  # Maximum statements allowed in functions
MAX_FUNCTION_BRANCHES = 12  # Maximum branches allowed in functions
MAX_FUNCTION_COMPLEXITY = 10  # Maximum cyclomatic complexity

# Data processing constants
IQ_PAIR_SIZE = 2  # Size of I-Q measurement pair
MEASUREMENTS_PER_ITERATION = 2  # Number of measurements per iteration in converter

# Measurement validation thresholds
MAX_TIME_SCALE_WARNING = 1e-3  # Time scale above 1 ms triggers warning for qubit measurements
MAX_TIME_SCALE_ERROR = 1.0  # Time scale above 1 s is invalid for qubit coherence

# Time unit defaults for fitting
DEFAULT_DECAY_TIME_S = 1e-6  # Default decay time: 1 μs
MIN_DECAY_TIME_S = 1e-9  # Minimum decay time: 1 ns
DEFAULT_FREQUENCY_HZ = 1e3  # Default detuning frequency: 1 kHz
MAX_FREQUENCY_HZ = 1e6  # Maximum detuning frequency: ±1 MHz

# Oscillation fitting
MIN_POINTS_PER_OSCILLATION = 4  # Minimum data points per oscillation period
MIN_OSCILLATIONS_FOR_FIT = 2  # Minimum oscillations needed for reliable fit

# Unit conversions
SECONDS_TO_MICROSECONDS = 1e6  # Multiply seconds by this to get microseconds

# Project root search
MAX_PROJECT_ROOT_DEPTH = 50  # Maximum directory levels to search for project root

# Test constants
TEST_DECAY_TIME = 10.0  # Test decay time value in seconds
TEST_PLOT_SIZE = 5  # Test plot size
TEST_PLOT_HEIGHT = 4  # Test plot height
TEST_DPI = 300  # Test DPI value
TEST_ALPHA = 0.7  # Test alpha value
TEST_GRID_ALPHA = 0.3  # Test grid alpha value
TEST_FIT_POINTS = 100  # Test fit points value
TEST_DATA_LENGTH = 3  # Test data length for validation



"""Measurement analyzer."""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, cast

import h5py
import matplotlib.pyplot as plt
import numpy as np
from lmfit import Model, Parameters
from lmfit.models import ExponentialModel

from metsuperq.analysis.constants import (
    DEFAULT_DECAY_TIME_S,
    DEFAULT_FIT_POINTS,
    DEFAULT_FREQUENCY_HZ,
    DEFAULT_MEASUREMENT_TYPE,
    EXPECTED_DATASET_ROWS,
    HDF5_PATTERN,
    MAX_FREQUENCY_HZ,
    MAX_TIME_SCALE_ERROR,
    MAX_TIME_SCALE_WARNING,
    MEASUREMENT_TYPES,
    MIN_DECAY_TIME_S,
    MIN_OSCILLATIONS_FOR_FIT,
    MIN_POINTS_FOR_FFT,
    MIN_POINTS_PER_OSCILLATION,
    PLOT_CONFIG,
    SECONDS_TO_MICROSECONDS,
)
from metsuperq.analysis.physics_validation import (
    validate_coherence_times,
    validate_ramsey_parameters,
)
from metsuperq.analysis.utils import get_project_root, save_analysis_results, save_summary_report
from metsuperq.utils import setup_logging

logger = setup_logging(__name__)

# Type aliases
AnalysisResult = dict[str, Any]
FitResult = dict[str, Any]
DatasetDict = dict[str, np.ndarray]
MetadataDict = dict[str, Any]


@dataclass
class RamseyParams:
    """Parameters for Ramsey decay function."""

    amp: float
    tau: float
    freq: float
    phi: float
    base: float


class BaseAnalyzer:
    """Measurement analyzer.

    Provides a unified interface for:
    - Finding and loading HDF5 measurement files
    - Exponential decay fitting for T1, T2 echo, T2 Ramsey measurements
    - Visualization and results export
    - Complete analysis pipeline from TUID to results
    """

    def __init__(self, data_path: Path | str | None = None) -> None:
        """Initialize analyzer with data directory path.

        Parameters
        ----------
        data_path : Path | str | None, optional
            Path to directory containing HDF5 files.
            If None, automatically detects project root and uses 'data' subdirectory.

        Attributes
        ----------
        data_path : Path
            Path to directory containing HDF5 files
        """
        if data_path is not None:
            self.data_path = Path(data_path)
        else:
            self.data_path = get_project_root() / "data"

        logger.info("Initialized BaseAnalyzer with data path: %s", self.data_path)

    def find_hdf5_file(self, tuid: str) -> Path | None:
        """Find HDF5 file based on TUID."""
        logger.debug("Searching for HDF5 file with TUID: %s", tuid)
        logger.debug("Searching in data path: %s", self.data_path)

        # First, check if data_path itself is a TUID directory
        if self.data_path.name == tuid:
            logger.debug("Data path matches TUID, searching directly in: %s", self.data_path)
            for hdf5_file in self.data_path.glob("*.hdf5"):
                logger.info("Found HDF5 file in TUID directory: %s", hdf5_file)
                return hdf5_file

        # Search for files containing the TUID in filename
        for hdf5_file in self.data_path.glob(HDF5_PATTERN):
            if tuid in hdf5_file.stem:
                logger.info("Found HDF5 file: %s", hdf5_file)
                return hdf5_file

        # Search in subdirectories named with the TUID
        tuid_subdir = self.data_path / tuid
        if tuid_subdir.exists() and tuid_subdir.is_dir():
            logger.debug("Searching in TUID subdirectory: %s", tuid_subdir)
            for hdf5_file in tuid_subdir.glob("*.hdf5"):
                logger.info("Found HDF5 file in TUID subdirectory: %s", hdf5_file)
                return hdf5_file

        logger.error("No HDF5 file found for TUID: %s", tuid)
        logger.error("Searched in: %s and subdirectories", self.data_path)
        return None

    def _build_dataset_name_map(self, subgroup_keys: Iterable[str]) -> dict[str, dict[str, str]]:
        """Build a map of base dataset names from HDF5 subgroup keys."""
        bases: dict[str, dict[str, str]] = {}
        for ds_name in subgroup_keys:
            if ds_name.endswith("_time"):
                base = ds_name[: -len("_time")]
                bases.setdefault(base, {})["time"] = ds_name
            elif ds_name.endswith("_iq"):
                base = ds_name[: -len("_iq")]
                bases.setdefault(base, {})["iq"] = ds_name
            else:
                bases.setdefault(ds_name, {})["combined"] = ds_name
        return bases

    def _load_combined_dataset(self, subgroup: h5py.Group, dataset_name: str) -> np.ndarray | None:
        """Load a combined dataset from HDF5 subgroup."""
        ds = subgroup[dataset_name]
        if isinstance(ds, h5py.Dataset):
            return ds[:]
        return None

    def _load_time_iq_pair(
        self, subgroup: h5py.Group, time_name: str, iq_name: str
    ) -> np.ndarray | None:
        """Load and combine time/IQ pair datasets."""
        ds_time = subgroup[time_name]
        ds_iq = subgroup[iq_name]
        if isinstance(ds_time, h5py.Dataset) and isinstance(ds_iq, h5py.Dataset):
            time_arr = np.asarray(ds_time[:])
            iq_arr = np.asarray(ds_iq[:])
            # Reconstruct into 2xN array (complex dtype for compatibility)
            return np.vstack(
                [
                    time_arr.astype(np.complex128),
                    iq_arr.astype(np.complex128),
                ]
            )
        return None

    def _process_subgroup_datasets(
        self, subgroup: h5py.Group, subgroup_name: str
    ) -> dict[str, np.ndarray]:
        """Process all datasets in a single HDF5 subgroup."""
        datasets: dict[str, np.ndarray] = {}
        # subgroup.keys() returns a KeysViewHDF5 which is iterable - no need for list()
        bases = self._build_dataset_name_map(subgroup.keys())

        for base, parts in bases.items():
            try:
                result = None
                if "combined" in parts:
                    result = self._load_combined_dataset(subgroup, parts["combined"])
                elif "time" in parts and "iq" in parts:
                    result = self._load_time_iq_pair(subgroup, parts["time"], parts["iq"])

                if result is not None:
                    datasets[f"{subgroup_name}_{base}"] = result
            except (OSError, KeyError, ValueError) as e:
                logger.warning(
                    "Failed to load dataset '%s' in subgroup '%s': %s", base, subgroup_name, e
                )
        return datasets

    def load_measurement_data(
        self, filepath: Path, measurement_type: str = DEFAULT_MEASUREMENT_TYPE
    ) -> tuple[MetadataDict, DatasetDict]:
        """Load measurement data and metadata from HDF5 file."""
        logger.debug("Loading data from file: %s", filepath)

        try:
            with h5py.File(filepath, "r") as f:
                if measurement_type not in f:
                    raise KeyError(f"Measurement type '{measurement_type}' not found")

                group = cast(h5py.Group, f[measurement_type])
                metadata = dict(group.attrs)
                datasets: dict[str, np.ndarray] = {}

                # Load all datasets from measurement iterations
                for subgroup_name in group.keys():
                    subgroup_item = group[subgroup_name]
                    if isinstance(subgroup_item, h5py.Group):
                        subgroup_datasets = self._process_subgroup_datasets(
                            subgroup_item, subgroup_name
                        )
                        datasets.update(subgroup_datasets)

                logger.debug(
                    "Loaded %d datasets with %d metadata fields", len(datasets), len(metadata)
                )
                return metadata, datasets

        except Exception as e:
            logger.error("Error loading data from %s: %s", filepath, e, exc_info=True)
            raise

    def _extract_time_amplitude_data(self, dataset: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """Extract time delays and amplitude data from complex dataset.

        Parameters
        ----------
        dataset : np.ndarray
            Complex dataset with shape (2, N) where first row is time delays
            and second row is complex I-Q measurements

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            Time delays and amplitude magnitudes
        """
        if dataset.shape[0] != EXPECTED_DATASET_ROWS:
            raise ValueError(
                f"Expected dataset with {EXPECTED_DATASET_ROWS} rows, got {dataset.shape[0]}"
            )

        # First row contains time delays (real part only)
        time_delays = np.real(dataset[0, :])

        # Second row contains complex I-Q measurements
        iq_data = dataset[1, :]
        amplitudes = np.abs(iq_data)

        logger.debug(
            "Extracted %d time points from %.2f to %.2f",
            len(time_delays),
            time_delays.min(),
            time_delays.max(),
        )

        return time_delays, amplitudes

    def _ramsey_decay_func(self, t: np.ndarray, params: RamseyParams) -> np.ndarray:
        """T2-Ramsey decaying sine function: A * exp(-t/T2*) * cos(2πft + φ) + B.

        Parameters
        ----------
        t : np.ndarray
            Time values
        params : RamseyParams
            Ramsey fit parameters (amp, tau, freq, phi, base)

        Returns
        -------
        np.ndarray
            Function values
        """
        return (
            params.amp * np.exp(-t / params.tau) * np.cos(2 * np.pi * params.freq * t + params.phi)
            + params.base
        )

    def _ramsey_decay_lmfit(self, t: np.ndarray, **kwargs: float) -> np.ndarray:
        """Provide fit wrapper for lmfit Model that uses individual parameters."""
        params = RamseyParams(
            amp=kwargs["amp"],
            tau=kwargs["tau"],
            freq=kwargs["freq"],
            phi=kwargs["phi"],
            base=kwargs["base"],
        )
        return self._ramsey_decay_func(t, params)

    def fit_exponential_decay(
        self, time_data: np.ndarray, amplitude_data: np.ndarray, measurement_type: str = "T1"
    ) -> FitResult:
        """Fit exponential decay model to measurement data.

        Parameters
        ----------
        time_data : np.ndarray
            Time delay values
        amplitude_data : np.ndarray
            Amplitude measurements
        measurement_type : str, optional
            Type of measurement for logging, by default "T1"

        Returns
        -------
        FitResult
            Dictionary containing fit parameters and uncertainties
        """
        logger.debug("Fitting exponential decay for %s measurement", measurement_type)

        # Create exponential decay model (model: y = amplitude * exp(-x/decay))
        model = ExponentialModel()

        # Set initial parameter guesses
        params = model.make_params()
        params["amplitude"].set(value=float(np.max(amplitude_data)), min=0)
        # lmfit ExponentialModel uses exp(-x/decay) where 'decay' is the time constant τ
        positive_times = time_data[time_data > 0]
        if positive_times.size > 0:
            # Use a reasonable fraction of the time range as initial guess
            init_tau = float(np.max(positive_times)) / 3  # Conservative estimate
        else:
            init_tau = DEFAULT_DECAY_TIME_S
        params["decay"].set(value=init_tau, min=MIN_DECAY_TIME_S)

        try:
            # Perform fit
            result = model.fit(amplitude_data, params, x=time_data)

            # Extract fit parameters and uncertainties. 'decay' is the time constant τ
            amp = float(result.params["amplitude"].value)
            amp_stderr = float(result.params["amplitude"].stderr or 0.0)
            decay_time = float(result.params["decay"].value)  # This is τ directly
            decay_time_stderr = float(result.params["decay"].stderr or 0.0)

            # Calculate decay rate for backward compatibility
            if decay_time > 0:
                decay_rate = 1.0 / decay_time
                # Error propagation: k = 1/τ => sigma_k = sigma_τ / τ^2
                decay_rate_stderr = (
                    decay_time_stderr / (decay_time**2) if decay_time_stderr else 0.0
                )
            else:
                decay_rate = float("inf")
                decay_rate_stderr = 0.0

            var_y = float(np.var(amplitude_data))
            if result.residual is not None and var_y > 0:
                r_squared = 1.0 - float(np.var(result.residual)) / var_y
            else:
                r_squared = 0.0

            fit_result = {
                "amplitude": amp,
                "amplitude_stderr": amp_stderr,
                # Keep backward-compatible keys while also storing rate
                "decay_time": decay_time,
                "decay_time_stderr": decay_time_stderr,
                "decay_rate": decay_rate,
                "decay_rate_stderr": decay_rate_stderr,
                "chi_squared": float(result.chisqr) if result.chisqr is not None else 0.0,
                "reduced_chi_squared": float(result.redchi) if result.redchi is not None else 0.0,
                "r_squared": r_squared,
                "fit_success": bool(result.success),
                "fit_type": "exponential",
            }

            logger.info(
                "%s fit: decay_time=%.1f±%.1f μs, amplitude=%.3f±%.3f",
                measurement_type,
                decay_time * SECONDS_TO_MICROSECONDS,
                decay_time_stderr * SECONDS_TO_MICROSECONDS,
                amp,
                amp_stderr,
            )

            return fit_result

        except Exception as e:
            logger.error("Fit failed for %s: %s", measurement_type, e, exc_info=True)
            raise

    def _validate_ramsey_data(self, time_data: np.ndarray) -> None:
        """Validate data for Ramsey fitting."""
        if len(time_data) < MIN_POINTS_FOR_FFT:
            raise ValueError(
                f"Insufficient data points for Ramsey fit: {len(time_data)} < {MIN_POINTS_FOR_FFT}"
            )

        max_time = np.max(time_data)
        if max_time > MAX_TIME_SCALE_WARNING:  # > 1 millisecond
            logger.warning(
                "Suspiciously large time scale for T2R: max_time=%.2e s. Expected µs scale.",
                max_time,
            )
        if max_time > MAX_TIME_SCALE_ERROR:
            raise ValueError(
                f"Time scale too large for qubit coherence: {max_time:.2e} s. Check units!"
            )

        if len(time_data) < MIN_POINTS_PER_OSCILLATION * MIN_OSCILLATIONS_FOR_FIT:
            logger.warning("Very few points for oscillating fit: %d points", len(time_data))

    def _estimate_ramsey_frequency(
        self, time_data: np.ndarray, amplitude_data: np.ndarray
    ) -> float:
        """Estimate detuning frequency using FFT."""
        if len(time_data) <= MIN_POINTS_FOR_FFT:
            return DEFAULT_FREQUENCY_HZ

        try:
            dt = np.mean(np.diff(time_data))
            freqs = np.fft.fftfreq(len(amplitude_data), dt)
            fft_vals = np.abs(np.fft.fft(amplitude_data - np.mean(amplitude_data)))
            # Find peak frequency (ignore DC component)
            peak_idx = np.argmax(fft_vals[1 : len(fft_vals) // 2]) + 1
            return abs(freqs[peak_idx])
        except (ValueError, FloatingPointError, ZeroDivisionError, IndexError):
            return DEFAULT_FREQUENCY_HZ

    def _setup_ramsey_parameters(
        self, time_data: np.ndarray, amplitude_data: np.ndarray
    ) -> Parameters:
        """Set up initial parameters for Ramsey fitting."""
        params = Parameters()

        # Amplitude: half the range of data (since it oscillates around baseline)
        amp_guess = (np.max(amplitude_data) - np.min(amplitude_data)) / 2
        params.add("amp", value=amp_guess, min=0)

        # T2* time: rough estimate from time range
        positive_times = time_data[time_data > 0]
        if positive_times.size > 0:
            tau_guess = float(np.max(positive_times)) / 3  # Conservative estimate
        else:
            tau_guess = DEFAULT_DECAY_TIME_S
        params.add("tau", value=tau_guess, min=MIN_DECAY_TIME_S)

        # Detuning frequency: estimate from oscillation period
        freq_guess = self._estimate_ramsey_frequency(time_data, amplitude_data)
        params.add("freq", value=freq_guess, min=-MAX_FREQUENCY_HZ, max=MAX_FREQUENCY_HZ)

        # Phase: start at 0
        params.add("phi", value=0.0, min=-np.pi, max=np.pi)

        # Baseline: mean of data
        base_guess = np.mean(amplitude_data)
        params.add("base", value=base_guess)

        return params

    def _extract_ramsey_fit_results(self, result, amplitude_data: np.ndarray) -> FitResult:
        """Extract fit results from lmfit result object."""
        # Extract fit parameters and uncertainties
        amp = float(result.params["amp"].value)
        amp_stderr = float(result.params["amp"].stderr or 0.0)
        tau = float(result.params["tau"].value)
        tau_stderr = float(result.params["tau"].stderr or 0.0)
        freq = float(result.params["freq"].value)
        freq_stderr = float(result.params["freq"].stderr or 0.0)
        phi = float(result.params["phi"].value)
        phi_stderr = float(result.params["phi"].stderr or 0.0)
        base = float(result.params["base"].value)
        base_stderr = float(result.params["base"].stderr or 0.0)

        var_y = float(np.var(amplitude_data))
        if result.residual is not None and var_y > 0:
            r_squared = 1.0 - float(np.var(result.residual)) / var_y
        else:
            r_squared = 0.0

        return {
            "amplitude": amp,
            "amplitude_stderr": amp_stderr,
            "decay_time": tau,  # T2* time
            "decay_time_stderr": tau_stderr,
            "frequency": freq,  # Detuning frequency
            "frequency_stderr": freq_stderr,
            "phase": phi,
            "phase_stderr": phi_stderr,
            "baseline": base,
            "baseline_stderr": base_stderr,
            "chi_squared": float(result.chisqr) if result.chisqr is not None else 0.0,
            "reduced_chi_squared": float(result.redchi) if result.redchi is not None else 0.0,
            "r_squared": r_squared,
            "fit_success": bool(result.success),
            "fit_type": "ramsey",
        }

    def fit_ramsey_decay(
        self, time_data: np.ndarray, amplitude_data: np.ndarray, measurement_type: str = "T2R"
    ) -> FitResult:
        """Fit T2-Ramsey decaying sine model to measurement data.

        Parameters
        ----------
        time_data : np.ndarray
            Time delay values
        amplitude_data : np.ndarray
            Amplitude measurements
        measurement_type : str, optional
            Type of measurement for logging, by default "T2R"

        Returns
        -------
        FitResult
            Dictionary containing fit parameters and uncertainties
        """
        logger.debug("Fitting T2-Ramsey decaying sine for %s measurement", measurement_type)

        # Validate data
        self._validate_ramsey_data(time_data)

        # Create custom Ramsey model
        model = Model(self._ramsey_decay_lmfit)

        # Set up parameters
        params = self._setup_ramsey_parameters(time_data, amplitude_data)

        try:
            # Perform fit
            result = model.fit(amplitude_data, params, t=time_data)

            # Extract fit parameters and uncertainties
            fit_result = self._extract_ramsey_fit_results(result, amplitude_data)

            # Validate Ramsey parameters for physical reasonableness
            param_warnings = validate_ramsey_parameters(
                frequency=fit_result["frequency"],
                phase=fit_result["phase"],
                amplitude=fit_result["amplitude"],
                baseline=fit_result["baseline"],
            )
            if param_warnings:
                fit_result["parameter_warnings"] = param_warnings
                for warning in param_warnings:
                    logger.warning("Ramsey parameter validation: %s", warning)

            logger.info(
                "%s fit: T2*=%.3f±%.3f μs, freq=%.1f±%.1f Hz, amp=%.3f±%.3f",
                measurement_type,
                fit_result["decay_time"] * 1e6,  # Convert to μs
                fit_result["decay_time_stderr"] * 1e6,
                fit_result["frequency"],
                fit_result["frequency_stderr"],
                fit_result["amplitude"],
                fit_result["amplitude_stderr"],
            )

            return fit_result

        except Exception as e:
            logger.error("Ramsey fit failed for %s: %s", measurement_type, e, exc_info=True)
            raise

    def analyze_coherence_data(self, filepath: Path) -> AnalysisResult:  # noqa: C901, PLR0912
        """Analyze coherence measurement data (T1, T2 echo, T2 Ramsey).

        Parameters
        ----------
        filepath : Path
            Path to HDF5 file containing coherence data

        Returns
        -------
        AnalysisResult
            Complete analysis results including fits and derived quantities
        """
        logger.info("Starting coherence analysis for file: %s", filepath)

        # Load data
        metadata, datasets = self.load_measurement_data(filepath)

        fits: dict[str, FitResult] = {}
        derived_quantities: dict[str, Any] = {}
        analysis_result: AnalysisResult = {
            "filepath": str(filepath),
            "metadata": metadata,
            "fits": fits,
            "derived_quantities": derived_quantities,
        }

        # Analyze each measurement type
        measurement_types = MEASUREMENT_TYPES

        for meas_type in measurement_types:
            # Find datasets for this measurement type
            matching_datasets = [k for k in datasets.keys() if meas_type in k.lower()]

            if not matching_datasets:
                logger.warning("No datasets found for measurement type: %s", meas_type)
                continue

            # Combine data from all measurement iterations
            all_times = []
            all_amplitudes = []

            for dataset_name in matching_datasets:
                dataset = datasets[dataset_name]
                times, amplitudes = self._extract_time_amplitude_data(dataset)
                all_times.extend(times)
                all_amplitudes.extend(amplitudes)

            if len(all_times) == 0:
                continue

            # Sort by time
            sorted_indices = np.argsort(all_times)
            sorted_times = np.array(all_times)[sorted_indices]
            sorted_amplitudes = np.array(all_amplitudes)[sorted_indices]

            # Choose appropriate fit based on measurement type
            try:
                if meas_type == "t2r":  # T2-Ramsey needs decaying sine fit
                    fit_result = self.fit_ramsey_decay(
                        sorted_times, sorted_amplitudes, meas_type.upper()
                    )
                else:  # T1 and T2E use exponential decay fit
                    fit_result = self.fit_exponential_decay(
                        sorted_times, sorted_amplitudes, meas_type.upper()
                    )
                fits[meas_type] = fit_result

                # Store derived quantities
                if meas_type == "t1":
                    derived_quantities["T1"] = fit_result["decay_time"]
                    derived_quantities["T1_stderr"] = fit_result["decay_time_stderr"]
                elif meas_type == "t2e":
                    derived_quantities["T2_echo"] = fit_result["decay_time"]
                    derived_quantities["T2_echo_stderr"] = fit_result["decay_time_stderr"]
                elif meas_type == "t2r":
                    derived_quantities["T2_ramsey"] = fit_result["decay_time"]
                    derived_quantities["T2_ramsey_stderr"] = fit_result["decay_time_stderr"]

            except (ValueError, RuntimeError) as e:
                logger.error("Failed to analyze %s data: %s", meas_type, e)
                continue

        logger.info("Coherence analysis completed with %d successful fits", len(fits))

        # Validate physics constraints (T2 ≤ 2×T1)
        validation = validate_coherence_times(analysis_result, strict=False)
        analysis_result["physics_validation"] = {
            "is_valid": validation.is_valid,
            "violations": validation.violations,
            "warnings": validation.warnings,
        }

        if not validation.is_valid:
            logger.error(
                "Physics validation FAILED with %d violation(s). Results may be unreliable!",
                len(validation.violations),
            )
            for violation in validation.violations:
                logger.error("  - %s", violation)
        elif validation.warnings:
            logger.warning("Physics validation passed with %d warning(s)", len(validation.warnings))
            for warning in validation.warnings:
                logger.warning("  - %s", warning)
        else:
            logger.info("Physics validation PASSED - all coherence times are physically valid")

        return analysis_result

    def _generate_fit_curve(
        self,
        fit_result: FitResult,
        time_range: tuple[float, float],
        num_points: int = DEFAULT_FIT_POINTS,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Generate fitted curve data for plotting.

        Parameters
        ----------
        fit_result : FitResult
            Fit results from fit_exponential_decay or fit_ramsey_decay
        time_range : tuple[float, float]
            Min and max time values for curve
        num_points : int, optional
            Number of points in curve, by default 100

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            Time points and fitted amplitude values
        """
        if not fit_result["fit_success"]:
            raise ValueError("Cannot generate curve from unsuccessful fit")

        x_fit = np.linspace(time_range[0], time_range[1], num_points)

        # Generate curve based on fit type
        if fit_result.get("fit_type") == "ramsey":
            # T2-Ramsey: A * exp(-t/T2*) * cos(2πft + φ) + B
            ramsey_params = RamseyParams(
                amp=fit_result["amplitude"],
                tau=fit_result["decay_time"],
                freq=fit_result["frequency"],
                phi=fit_result["phase"],
                base=fit_result["baseline"],
            )
            y_fit = self._ramsey_decay_func(x_fit, ramsey_params)
        else:
            # Exponential: A * exp(-t/τ)
            y_fit = fit_result["amplitude"] * np.exp(-x_fit / fit_result["decay_time"])

        return x_fit, y_fit

    def _create_empty_plot_if_needed(self, save_path: Path | None) -> bool:
        """Create empty plot if needed and return True if handled."""
        if save_path:
            _fig, ax = plt.subplots(
                1, 1, figsize=(PLOT_CONFIG["figsize_per_plot"], PLOT_CONFIG["height"])
            )
            ax.set_title("No fits available")
            ax.axis("off")
            plt.tight_layout()
            plt.savefig(save_path, dpi=PLOT_CONFIG["dpi"], bbox_inches="tight")
            plt.close()
        return True

    def _collect_measurement_data(
        self, datasets: DatasetDict, meas_type: str
    ) -> tuple[np.ndarray, np.ndarray]:
        """Collect and sort time/amplitude data for a measurement type."""
        matching_datasets = [k for k in datasets.keys() if meas_type in k.lower()]

        all_times = []
        all_amplitudes = []

        for dataset_name in matching_datasets:
            dataset = datasets[dataset_name]
            times, amplitudes = self._extract_time_amplitude_data(dataset)
            all_times.extend(times)
            all_amplitudes.extend(amplitudes)

        # Sort data
        sorted_indices = np.argsort(all_times)
        sorted_times = np.array(all_times)[sorted_indices]
        sorted_amplitudes = np.array(all_amplitudes)[sorted_indices]

        return sorted_times, sorted_amplitudes

    def _create_fit_label(self, fit_result: FitResult) -> str:
        """Create appropriate label based on fit type."""
        if fit_result.get("fit_type") == "ramsey":
            decay_time_us = fit_result["decay_time"] * 1e6
            decay_stderr_us = fit_result["decay_time_stderr"] * 1e6
            freq = fit_result["frequency"]
            freq_stderr = fit_result["frequency_stderr"]
            return (
                f"Fit: T₂*={decay_time_us:.1f}±{decay_stderr_us:.1f} μs, "
                f"f={freq:.0f}±{freq_stderr:.0f} Hz"
            )
        else:
            # Convert to microseconds for better readability
            decay_us = fit_result["decay_time"] * 1e6
            decay_stderr_us = fit_result["decay_time_stderr"] * 1e6
            return f"Fit: τ={decay_us:.1f}±{decay_stderr_us:.1f} μs"

    def plot_analysis(self, analysis_result: AnalysisResult, save_path: Path | None = None) -> None:
        """Plot analysis results with fits.

        Parameters
        ----------
        analysis_result : AnalysisResult
            Analysis results from analyze_coherence_data
        save_path : Path | None, optional
            Path to save plot PNG file, by default None
        """
        logger.debug("Creating analysis plots")

        # Load original data for plotting
        filepath = Path(analysis_result["filepath"])
        _, datasets = self.load_measurement_data(filepath)

        fits = analysis_result.get("fits", {})
        n_plots = len([mt for mt in MEASUREMENT_TYPES if mt in fits])

        if n_plots == 0:
            logger.warning("No successful fits to plot")
            self._create_empty_plot_if_needed(save_path)
            return

        _fig, axes = plt.subplots(
            1, n_plots, figsize=(PLOT_CONFIG["figsize_per_plot"] * n_plots, PLOT_CONFIG["height"])
        )
        if n_plots == 1:
            axes = [axes]

        plot_idx = 0

        for meas_type in MEASUREMENT_TYPES:
            if meas_type not in fits:
                continue

            ax = axes[plot_idx]
            sorted_times, sorted_amplitudes = self._collect_measurement_data(datasets, meas_type)

            # Plot data points
            ax.scatter(sorted_times, sorted_amplitudes, alpha=PLOT_CONFIG["alpha"], label="Data")

            # Plot fit if successful
            fit_result = fits[meas_type]
            if fit_result["fit_success"]:
                time_range = (sorted_times.min(), sorted_times.max())
                x_fit, y_fit = self._generate_fit_curve(fit_result, time_range)
                label = self._create_fit_label(fit_result)
                ax.plot(x_fit, y_fit, "r-", label=label)

            ax.set_xlabel("Time")
            ax.set_ylabel("Amplitude")
            ax.set_title(f"{meas_type.upper()} Measurement")
            ax.legend()
            ax.grid(True, alpha=PLOT_CONFIG["grid_alpha"])

            plot_idx += 1

        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, dpi=PLOT_CONFIG["dpi"], bbox_inches="tight")
            logger.info("Plot saved to: %s", save_path)
        else:
            plt.show()

        plt.close()

    def save_analysis_results(self, analysis_result: AnalysisResult, output_path: Path) -> None:
        """Save analysis results to JSON file.

        Parameters
        ----------
        analysis_result : AnalysisResult
            Analysis results to save
        output_path : Path
            Path for output JSON file
        """
        save_analysis_results(analysis_result, output_path)

    def save_summary_report(self, analysis_result: AnalysisResult, output_path: Path) -> None:
        """Save a text summary report to file.

        Parameters
        ----------
        analysis_result : AnalysisResult
            Analysis results to summarize
        output_path : Path
            Path for output text file
        """
        save_summary_report(analysis_result, output_path)

    def run_full_analysis(self, tuid: str, output_dir: Path | None = None) -> AnalysisResult:
        """Run complete analysis pipeline for a given TUID.

        Parameters
        ----------
        tuid : str
            Timestamp-based unique identifier for the measurement
            (format: YYYYmmDD-HHMMSS-sss-******)
        output_dir : Path | None, optional
            Directory for output files, by default None (uses data_path)

        Returns
        -------
        AnalysisResult
            Complete analysis results
        """
        logger.info("Starting full analysis for TUID: %s", tuid)

        # Find HDF5 file
        hdf5_file = self.find_hdf5_file(tuid)
        if hdf5_file is None:
            raise FileNotFoundError(f"No HDF5 file found for TUID: {tuid}")

        # Run analysis
        analysis_result = self.analyze_coherence_data(hdf5_file)

        # Set up output directory
        if output_dir is None:
            output_dir = self.data_path / tuid  # Save to TUID-specific folder
        output_dir = Path(output_dir)
        output_dir.mkdir(exist_ok=True)

        # Save results
        json_path = output_dir / f"{tuid}_analysis_results.json"
        self.save_analysis_results(analysis_result, json_path)

        # Create and save plot
        plot_path = output_dir / f"{tuid}_analysis_plot.png"
        self.plot_analysis(analysis_result, plot_path)

        # Save summary report
        report_path = output_dir / f"{tuid}_summary_report.txt"
        self.save_summary_report(analysis_result, report_path)

        logger.info("Full analysis completed for TUID: %s", tuid)
        return analysis_result



"""Database integration module for InfluxDB v2."""

from __future__ import annotations

from metsuperq.database.client import InfluxDBClient
from metsuperq.database.config import InfluxConfig, InfluxConfigError
from metsuperq.database.exceptions import (
    DatabaseError,
    InfluxDBConnectionError,
    ValidationError,
    WriteError,
)
from metsuperq.database.schema import (
    AnalysisMetadataFields,
    AnalysisMetadataPoint,
    AnalysisMetadataTags,
    FitQuality,
    InfluxPoint,
    MeasurementPoint,
    MeasurementTags,
    MeasurementType,
    T1MeasurementFields,
    T2EchoMeasurementFields,
    T2RamseyMeasurementFields,
)
from metsuperq.database.writer import MeasurementWriter

__all__ = [
    "InfluxConfig",
    "InfluxConfigError",
    "InfluxDBClient",
    "MeasurementWriter",
    "DatabaseError",
    "InfluxDBConnectionError",
    "ValidationError",
    "WriteError",
    "MeasurementPoint",
    "MeasurementTags",
    "MeasurementType",
    "FitQuality",
    "InfluxPoint",
    "AnalysisMetadataFields",
    "AnalysisMetadataPoint",
    "AnalysisMetadataTags",
    "T1MeasurementFields",
    "T2EchoMeasurementFields",
    "T2RamseyMeasurementFields",
]
